#include "Cache.hpp"

/******************************************************************************
 * BlackFish
 * Copyright (C) 1999-2001 Sren Meyer-Eppler, Jutta Jordans
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ******************************************************************************/

//-------------------------------------------------------------------------------------------------
///    Notes:
///    - order the if() statements to minimize jumps
///    - do I really need a doubly linked list instead of a singly linked one?
///    - I need an update() function to sync the cached in surface with the
///      possibly modified source surface, otherwise the cache might have an older
///      version of the surface and keep blitting that
//-------------------------------------------------------------------------------------------------

#pragma warning (push, 1)
#include <cassert>
#pragma warning (pop)

#include "General/HelpMacros.hpp"
#include "BlackFish/DxDraw.hpp"
#include "BlackFish/Surface.hpp"
#include "BlackFish/Paint.hpp"


//-------------------------------------------------------------------------------------------------


/**
 * initializes a new slot, slots are the atomic unit the cachesurfaces are
 * devided in, i.e. if you cache in a new surface it occupies exactly one slot.
 * the class is a simple data wrapper, nothing is allocated, no member functions
 */
CCache::cPOOL::cSLOT::cSLOT(
	const RECT &Rect,
    CSurface *CacheSurface,
    CSurface *SourceSurface,
	cPOOL *Pool,
	cSLOT *Prev,
	cSLOT *Next )
	: Rect( Rect )
	 ,CacheSurface( CacheSurface )
	 ,SourceSurface( SourceSurface )
	 ,Pool( Pool )
	 ,Prev( Prev )
	 ,Next( Next )
{
}


//-------------------------------------------------------------------------------------------------


/**
 * a pool is a collection of cache surfaces for a particular size of slots.
 * i.e. we have a pool for 64x32 slots, a pool for 64x128 slots etc...
 */
CCache::cPOOL::cPOOL( tSINT32 SlotWidth, tSINT32 SlotHeight, tSINT32 NrOfSurfaces, tUINT32 TransparentColor )
	: SlotWidth( SlotWidth ),
	  SlotHeight( SlotHeight ),
	  NrOfFreeSlots( NrOfSurfaces * SurfaceWidth / SlotWidth * SurfaceWidth / SlotHeight ),
	  First( NULL ),
	  Last( First ),
	  CacheSurface( NrOfSurfaces ),
  	  OverflowCount( 0 )
{
	assert( DirectDraw != NULL );
	assert( NrOfSurfaces > 0 );
	assert( SurfaceWidth % SlotWidth == 0 );
	assert( SurfaceWidth % SlotHeight == 0 );

    CPaint Painter;
	cSLOT *Iterator = NULL;
	cSLOT *Temp		= NULL;

	tSINT32 SlotsPerRow		= SurfaceWidth / SlotWidth;

    for ( std::vector< CSurface * >::iterator i = CacheSurface.begin( ); i != CacheSurface.end( ); ++i )
	{
        *i = new CSurface( );
		assert( *i != NULL );

        (*i)->Create( SurfaceWidth, SurfaceWidth, true );  // square and in vidmem
		(*i)->SetTransparentColor( TransparentColor );
		Painter.SetDrawingTarget( *i );
		Painter.Fill( 0 );	// clear to black

		for ( tSINT32 j = 0; j < NrOfFreeSlots / NrOfSurfaces; ++j )// for each slot in surface
		{
			RECT Rect;
			Rect.left	= (j % SlotsPerRow) * SlotWidth;
			Rect.top	= (j / SlotsPerRow) * SlotHeight;
			Rect.right	= Rect.left + SlotWidth;
			Rect.bottom	= Rect.top + SlotHeight;

			Temp = new cSLOT( Rect, *i, NULL, this, Iterator, NULL );
			assert( Temp != NULL );

			if ( Iterator != NULL )
				Iterator->Next = Temp;
			else					// Iterator == NULL => first node
				First = Temp;

			Iterator = Temp;
		}
	}

	Last = Iterator;
	assert( Last != NULL );
	assert( Last->Next == NULL );
}


CCache::cPOOL::~cPOOL( )
{
    for ( std::vector< CSurface * >::iterator i = CacheSurface.begin( ); i != CacheSurface.end( ); ++i )
		SAFEDELETE( *i );

	cSLOT *Iterator = Last;
	cSLOT *Temp;
	while ( Iterator != NULL )
	{
		Temp	 = Iterator;
		Iterator = Iterator->Prev;
		SAFEDELETE( Temp );
	}

	First = NULL;
	Last  = NULL;

*LOG << "Cache " << SlotWidth << "x" << SlotHeight << " has overflown " << OverflowCount << " times." << std::endl;
}


void
CCache::cPOOL::Invalidate( void )
{
    cSLOT *Temp = First;
    while ( Temp != NULL )
    {
        if ( Temp->SourceSurface != NULL )
            const_cast<CSurface *>( Temp->SourceSurface )->SetCacheStatus( NULL );
        Temp->SourceSurface = NULL;
        Temp = Temp->Next;
		NrOfFreeSlots++;
    }
}


/**
 * make this slot the most recently used one
 */
void
CCache::cPOOL::PromoteSlot( cSLOT *Slot )
{
    assert( Slot != NULL );
	assert( First != NULL );

    cSLOT *TempSlot1 = NULL;
    cSLOT *TempSlot2 = NULL;

    if ( Slot == First )                // resorting the list, current surface will be the first in the list
    {                                   // current surface already is the first, no need to sort
    }                                   // note: this takes care of the first == last issue, too
    else if ( Slot == Last )            // current surface is the last
    {                                   // insert current last Slot just before the first Slot and adjust
        TempSlot1       = Last->Prev;   // the first and last pointers accordingly
        TempSlot1->Next = NULL;
        Last->Next      = First;
        Last->Prev      = NULL;
        First->Prev     = Last;
        First           = Last;
        Last            = TempSlot1;
    }
    else                                // surface is neither first nor last in the list
    {                                   // remove current Slot from the list and insert it before the first
        TempSlot1       = Slot->Prev;   // adjust first pointer
        TempSlot2       = Slot->Next;
        TempSlot1->Next = TempSlot2;
        TempSlot2->Prev = TempSlot1;
        Slot->Prev      = NULL;
        Slot->Next      = First;
        First->Prev     = Slot;
        First           = Slot;
    }
}


/**
 * cache in a new surface
 */
CCache::cPOOL::cSLOT *
CCache::cPOOL::CacheInSurface( const CSurface *Surface )
{
    assert( Surface != NULL );
	assert( First != NULL );
	assert( Last != NULL );

    bool Found = false;
    cSLOT *Temp = NULL;

    RECT TempRect = Surface->GetSize();

    if ( NrOfFreeSlots > 0 )   // we have some free cache space left for the new surface
    {
        NrOfFreeSlots--;
        Temp  = First;
        Found = false;
        while ( !Found )            // find a slot that isn't used, yet
        {
            if ( Temp->SourceSurface == NULL )  // no source surface for this rect -> rect is unused
            {
                Found = true;
            }
			else
            {
                Temp = Temp->Next;  // haven't found a suitable rect, yet -> continue through the list
				assert( Temp != NULL );
            }
        }
    }
    else
    {
        // no CacheSurface has room left, discard oldest cached-in surface...
        const_cast<CSurface *>( Last->SourceSurface )->SetCacheStatus( NULL );
		Temp = Last;

		OverflowCount++;
    }

	OffsetRect( &TempRect, Temp->Rect.left, Temp->Rect.top );
    Temp->SourceSurface = Surface;   // !! cyclcic dependancy really necessary here?!
    const_cast<CSurface *>( Temp->SourceSurface )->SetCacheStatus( 
		reinterpret_cast<void *>( Temp )); // whuhui! const and reinterpret in one line! ;)
    Temp->CacheSurface->Blit( *Surface, NULL, &TempRect );

	return Temp;
}


//-------------------------------------------------------------------------------------------------


/**
 * creates the cache by initializing all pools
 */
CCache::CCache(
	tSINT32 NrOf16x16,		// !! these should be made percentage weights and the whole available
	tSINT32 NrOf64x32,		//    vidmem filled with surfaces according to the weights
	tSINT32 NrOf64x128,
	tSINT32 NrOf128x128,
    tUINT32 TransparentColor )
	: PoolNr16x16( 0 )
	 ,PoolNr64x32( 1 )
	 ,PoolNr64x128( 2 )
	 ,PoolNr128x128( 3 )
	 ,Pool( 4 )

{
	Pool[PoolNr16x16] = new cPOOL( 16, 16, NrOf16x16, TransparentColor );
	assert( Pool[PoolNr16x16] != NULL );

	Pool[PoolNr64x32] = new cPOOL( 64, 32, NrOf64x32, TransparentColor );
	assert( Pool[PoolNr64x32] != NULL );

	Pool[PoolNr64x128] = new cPOOL( 64, 128, NrOf64x128, TransparentColor );
	assert( Pool[PoolNr64x128] != NULL );

	Pool[PoolNr128x128] = new cPOOL( 128, 128, NrOf128x128, TransparentColor );
	assert( Pool[PoolNr128x128] != NULL );
}


CCache::~CCache( )
{
	for ( std::vector< cPOOL * >::iterator i = Pool.begin( ); i != Pool.end( ); ++i )
		SAFEDELETE( *i );
}


/**
 * removes all surfaces from the cache which causes them to be reloaded from systemmmory
 * usefull to recover from a corrupted cache, after a taskswitch for example
 */
void
CCache::Invalidate( void )
{
	for ( std::vector< cPOOL * >::iterator i = Pool.begin( ); i != Pool.end( ); ++i )
		(*i)->Invalidate( );
}


/**
 * removes a single surface from the cache. usefull when the original surface has
 * been modified and the copy in the cache is out of date
 */
void
CCache::Invalidate( CSurface *Surface )
{
	assert( Surface != NULL );

	cPOOL::cSLOT *Slot = reinterpret_cast<cPOOL::cSLOT *>( Surface->GetCacheStatus( ));
	if ( Slot != NULL )
	{
		Slot->Pool->NrOfFreeSlots++;
		Slot->SourceSurface = NULL;
		Surface->SetCacheStatus( NULL );
	}
	DEBUGCODE (
	else
	{
		*LOG << "Tried to invalidate a surface not in cache!" << std::endl;
	}
	)
}


/**
 * caches in a new surface
 * !! this could be optimized if each surface stored a pointer to the pool it belongs to
 *    the pointer would be initialized on first cache-in
 */
CCache::cPOOL::cSLOT *
CCache::CacheInSurface( const CSurface *Surface )
{
    assert( Surface != NULL );

    RECT TempRect = Surface->GetSize();

	tUINT32 PoolNr = 0;
	if ( TempRect.right <= 16 )
		PoolNr = PoolNr16x16;
	else if ( TempRect.right <= 64 )
		if ( TempRect.bottom <= 32 )
			PoolNr = PoolNr64x32;
		else if ( TempRect.bottom <= 128)
			PoolNr = PoolNr64x128;

	return Pool[PoolNr]->CacheInSurface( Surface );
}


/**
 * Checks if a surface is already in the cache, if so, it moves up in the priority
 * queue to be the last used one and gets blitted. If the requested surface is
 * not already in the cache it gets blitted from system memory into a cache surface
 * and then the desired blit to the target surface is done from cachesurface to destsurface
 * Note: the rects are allowed to be null, meaning the entire surface is to be blitted()
 */
void
CCache::Blit( 
	const CSurface &Source, 
	const RECT *SourceRect, 
	CSurface &Dest, 
	const RECT *DestRect )
{
	cPOOL::cSLOT *Slot = NULL;
    Slot = reinterpret_cast<cPOOL::cSLOT *>( Source.GetCacheStatus());

    // the surface is not yet in the cache
    if ( Slot == NULL )
		Slot = CacheInSurface( &Source );
    assert( Slot != NULL );

    Slot->Pool->PromoteSlot( Slot );

    RECT CacheSourceRect;
    if ( SourceRect != NULL)
    {
		CacheSourceRect = *SourceRect;
		OffsetRect( &CacheSourceRect, Slot->Rect.left, Slot->Rect.top );
    }
	else                              // the sourcerect is NULL, meaning that the user wants to blit
	{                                 // the entire surface
		CacheSourceRect = Slot->Rect;
    }

    assert( Slot->SourceSurface == &Source );
    Dest.Blit( *Slot->CacheSurface,  &CacheSourceRect, DestRect );  // finally: blit from cache
}


/**
 * this function is (and should be) an exact copy of CCache::Blit() except for the last
 * line, wich calls CSurface::Blit() or CSurface::BlitTrans() respectively
 * Possible ways to minimize code dublication:
 * - move common code into an extra function PrepareBlit() | I didn't want the overhead
 *   of an additional function call in an often called/important function such as this
 *   - 08.10.1999 did just this, see the functions MoveNodeUp() and CacheInSurface()
 * - Pass a flag to the function, telling it to either blittrans() or blit(), branching
 *   in an if() statement | I didn't want to break consistency with the current CSurface
 *   interface of having two seperate functions
 */
void
CCache::BlitTrans( 
	const CSurface &Source, 
	const RECT *SourceRect, 
	CSurface &Dest, 
	const RECT *DestRect )
{
	cPOOL::cSLOT *Slot = NULL;
    Slot = reinterpret_cast<cPOOL::cSLOT *>( Source.GetCacheStatus());

    // the surface is not yet in the cache
    if ( Slot == NULL )
		Slot = CacheInSurface( &Source );
    assert( Slot != NULL );

    Slot->Pool->PromoteSlot( Slot );

    RECT CacheSourceRect;
    if ( SourceRect != NULL)
    {
		CacheSourceRect = *SourceRect;
		OffsetRect( &CacheSourceRect, Slot->Rect.left, Slot->Rect.top );
    }
	else                              // the sourcerect is NULL, meaning that the user wants to blit
	{                                 // the entire surface
		CacheSourceRect = Slot->Rect;
    }

    assert( Slot->SourceSurface == &Source );
    Dest.BlitTrans( *Slot->CacheSurface,  &CacheSourceRect, DestRect );  // finally: blit from cache
}
