#include "Game.hpp"

/******************************************************************************
 * BlackFish
 * Copyright (C) 1999-2001 Sren Meyer-Eppler, Jutta Jordans
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ******************************************************************************/


#pragma warning ( push, 1 )
#include <stdlib.h>                     // rand, itoa
#pragma warning ( pop )

#include "general\debug.hpp"
#include "general\helpmacros.hpp"
#include "BlackFish\BlackFish.hpp"
#include "BlackFish\Cache.hpp"
#include "BlackFish\DXDraw.hpp"
#include "BlackFish\Input.hpp"
#include "BlackFish\Surface.hpp"
#include "BlackFish\PixelConverter.hpp"
#include "BlackFish\paint.hpp"
#include "BlackFish\TileSet.hpp"
#include "BlackFish\Time.hpp"
#include "BlackFish\write.hpp"
#include "BlackFish\sprite.hpp"


/*
 * base class, all other handlers must be derived from this
 * default constructor is initialized with the next handler in the scene
 */
CEventHandler::CEventHandler( CEventHandler *NextHandler = NULL )
	: NextHandler( NextHandler )
{
}

/*
 *
 */
CEventHandler::~CEventHandler( )
{
	NextHandler = NULL;
}

/*
 * base handler can't handle any events and just passes them along
 * this method is intended to be overwritten
 */
void
CEventHandler::HandleEvent( CEvent *Event )
{
	// default behaviour: can't deal with this event, just pass it along
	// note: if NextHandler is NULL, the Event drops off the end of the chain unhandled
	if ( NextHandler )
		NextHandler->HandleEvent( Event );
}

/*
 * handles all events relating to the map, i.e. mouseclicks in the mapwindow,
 * mousemovement (scrolling) etc
 */
CMapEventHandler::CMapEventHandler(
    CBlackFish *BlackFish,
    CBobbingSprite *MarineMarker,
    CSprite *CellMarker,
    CEventHandler *NextHandler )
    : CEventHandler( NextHandler )
 	 ,BlackFish( BlackFish )
	 ,MarineMarker( MarineMarker )
	 ,CellMarker( CellMarker )
	 ,SelectedMarine( NULL )
{
	ASSERT( MarineMarker != NULL );
	ASSERT( CellMarker   != NULL );
	ASSERT( BlackFish    != NULL );
}

/*
 * handles:
 * - left clicks on map-window
 * - mouse-movement at the screen-borders (scrolling)
 */
void
CMapEventHandler::HandleEvent( CEvent *Event )
{
	/**********************************/
	// try MouseClickEvent
    if ( CMouseClickEvent *MouseClickEvent = dynamic_cast<CMouseClickEvent *> ( Event ))
	{
		tSINT32 ArrayX, ArrayY;
		// leftclick
        if ((( MouseClickEvent->LastClick & CMouse::Button1 ) != 0 )
		   && ( BlackFish->GetCellForScreen( ArrayX, ArrayY, MouseClickEvent->PositionX, MouseClickEvent->PositionY )))
		{
			// kein Marine selektiert, checken, ob auf dem Feld einer steht
            std::list<CSprite *> *List = BlackFish->GetMap( )->GetSpriteList( ArrayX, ArrayY );
            std::list<CSprite *>::iterator Iterator;
			for ( Iterator = List->begin( ); Iterator != List->end( ); Iterator++ )
			{
				// Marine angeklickt
                if ( CMarineSprite *Marine = dynamic_cast<CMarineSprite *> ( *Iterator ))
				{
					if ( SelectedMarine != NULL )
						SelectedMarine->Detach( MarineMarker );
					SelectedMarine = Marine;
					MarineMarker->SetSubject( SelectedMarine );
					SelectedMarine->Attach( MarineMarker );
					break;
				}
				// irgendwas anderes getroffen
				else
				{
				}
			}

			// bereits ein Marine selektiert -> move
			if ( SelectedMarine != NULL )
                SelectedMarine->Move( CVector( static_cast<tFLOAT>( ArrayX * 32 ), static_cast<tFLOAT>( ArrayY * 32 ), 0.0f ));
		}
		// rightclick
        else if ((( MouseClickEvent->LastClick & CMouse::Button2 ) != 0 )
			&& ( SelectedMarine != NULL ))
		{
			SelectedMarine->Detach( MarineMarker );
			MarineMarker->SetSubject( NULL );
			SelectedMarine = NULL;
		}

        if ( Event->Persistant( )) CEventHandler::HandleEvent( Event );
		return;
	}

	/**********************************/
	// try MouseMoveEvent
    if ( CMouseMoveEvent *MouseMoveEvent = dynamic_cast<CMouseMoveEvent*> ( Event ))
	{
		BlackFish->Scroll( MouseMoveEvent->PositionX, MouseMoveEvent->PositionY );

		tSINT32 ArrayX, ArrayY;
		if ( BlackFish->GetCellForScreen( ArrayX, ArrayY, MouseMoveEvent->PositionX, MouseMoveEvent->PositionY ))
		{
			// highlight the mapcell the mouse cursor hovers above
            CellMarker->SetPosition( CVector( ArrayX * 32.0f, ArrayY * 32.0f, 0.0f ));
		}

        if ( Event->Persistant( )) CEventHandler::HandleEvent( Event );
		return;
	}

    CEventHandler::HandleEvent( Event );
}

/*
 * handles general keystrokes
 */
CMainKeyEventHandler::CMainKeyEventHandler(
	HWND nWindow,
    CEventHandler *nNextHandler )
    : CEventHandler( nNextHandler ),
	  Window( nWindow )
{
}

/*
 * handles:
 * - ESC key: quit program
 */
void
CMainKeyEventHandler::HandleEvent( CEvent *Event )
{
	/**********************************/
	// try KeyESCEvent
    if ( CKeyEscEvent *KeyESCEvent = dynamic_cast<CKeyEscEvent*> ( Event ))
	{
		PostMessage( Window, WM_CLOSE, 0, 0 );
        if ( Event->Persistant( )) CEventHandler::HandleEvent( Event );
		return;
	}

    CEventHandler::HandleEvent( Event );
}

//////////////////////////////////////////////

/*
 * draws miscellaneous info such as framerate, current position etc...
 */
void
CGame::ShowDebugInfo( void )
{
	static tSINT64 Frequency	= Timer.GetFrequency();
    static tSINT64 FrameTime	= 0;
	static tSINT64 OldFrameTime	= 0;
	static tDOUBLE FrameRate 	= 0.0;
    static tSINT64 Frames		= 0;

    Frames++;
    FrameTime = Timer.GetTime();
    if ( (FrameTime - OldFrameTime) >= Frequency )
    {
        FrameRate = (double)(Frames * Frequency) / (double)(FrameTime - OldFrameTime);
        OldFrameTime = Timer.GetTime();
        Frames = 0;
    }

    char MiscString[20];
    itoa( static_cast<tSINT32>( FrameRate ), MiscString, 10 );
    Writer->Out( 3, 0, "FPS:" );
    Writer->Out( 56, 0, MiscString );

    itoa( BlackFish->WorldX(), MiscString, 10);
    Writer->Out( 3,16, "WrldX:" );
    Writer->Out( 56,16, MiscString );

    itoa( BlackFish->WorldY(), MiscString, 10);
    Writer->Out( 3,32, "WrldY:" );
    Writer->Out( 56,32, MiscString );

    CVector Pos;
	Marine[0]->GetPosition( Pos );
    itoa( static_cast<int>( Pos.x ), MiscString, 10);
    Writer->Out( 3,48, "SprtX:" );
    Writer->Out( 56,48, MiscString );
    itoa( static_cast<int>( Pos.y ), MiscString, 10);
    Writer->Out( 3,64, "SprtY:" );
    Writer->Out( 56,64, MiscString );

	tSINT32 mx, my,w;
	tSINT32 x = 23, y = 42;
	tUINT32 b;
	Mouse->GetState( mx, my, w, b );
	if ( BlackFish->GetCellForScreen( x, y, mx, my ))
	{
		itoa( x, MiscString, 10);
		Writer->Out( 3,80, "CellX:" );
		Writer->Out( 56,80, MiscString );
		itoa( y, MiscString, 10);
		Writer->Out( 3,96, "CellY:" );
		Writer->Out( 56,96, MiscString );
	}
	else
	{
		Writer->Out( 3,80, "CellX:" );
		Writer->Out( 56,80, "-" );
		Writer->Out( 3,96, "CellY:" );
		Writer->Out( 56,96, "-" );
	}
}

/*
 * master setup, called once from InitApplication
 * important: not all objects can be properly set-up in the initializer list because
 *			  they unfortunately depend on global DirectDraw, which is setup just here...
 * important: very oder dependent! be careful with changes
 */
CGame::CGame( HWND WindowHandle, HINSTANCE Instance, bool FullScreen )
    : ApplicationActive( false )
	 ,Primary( NULL )
	 ,BackBuffer( NULL )
	 ,RenderTarget( NULL )
	 ,Font1( NULL )
     ,MouseCursor( NULL )
	 ,Cache( NULL )
     ,Painter( new CPaint( ))
     ,Input( new CInput( Instance ))
     ,KeyBoard( new CKeyboard( WindowHandle, Input ))
     ,Mouse( new CMouse( WindowHandle, Input ))
     ,BlackFish( NULL )
     ,MapEventHandler( NULL )
	 ,MainKeyEventHandler( NULL )
	 ,RootEventHandler( NULL )
	 ,Marine( )
	 ,TestBobber( NULL )
	 ,Marker( NULL )
	 ,NextFrameTime( 0 )
	 ,MaxFPS( 60 )	// !! should be read from an ini file?
{
*LOG << "Timer initialized with frequency: " << static_cast<tSINT32>( Timer.GetFrequency()) << std::endl;

    srand( static_cast<tUINT32>( Timer.GetTime( )));

	// gloabal DirectDraw object, MUST be initialized first
    DirectDraw  = new CDxDraw( WindowHandle, FullScreen );
	ASSERT( DirectDraw != NULL );

	// create Surfaces
    Primary     = new CSurface( );
	ASSERT( Primary != NULL );
    Primary->Create( );

    BackBuffer  = new CSurface( );
	ASSERT( BackBuffer != NULL );
	if ( DirectDraw->IsFullScreen( ))
		BackBuffer->Create( Primary ); // We're fullscreen, create a DX flipping chain
	else
        BackBuffer->Create( 800, 600, true ); // We're windowed, create offcreen vidmem backbuffer

	RenderTarget = BackBuffer;
    Painter->SetDrawingTarget( RenderTarget );

    CPixelConverter PixelConverter( RenderTarget );
    tUINT32 TransparentColor = PixelConverter.RGBToColor( 255, 0, 255 );	// remember pink transparent color

	// create cache
    Cache = new CCache( 1, 4, 4, 1, TransparentColor );
	ASSERT( Cache != NULL );

    // Load Game TileSets that are not stored in CBlackFish
    Font1       = new CTileSet( "graphics\\font.ini", "[Font1]" );
	ASSERT( Font1 != NULL );
    Font1->SetCache( Cache );

    MouseCursor = new CTileSet( "graphics\\mouse.ini", "[MouseCursor]" );
	ASSERT( MouseCursor != NULL );
    MouseCursor->SetCache( Cache );

	// create our Textout object
    Writer = new CWrite( );
	Writer->SetDrawingTarget( RenderTarget );
    Writer->SetFont( Font1 );
    Writer->SetSpacing( 0 );

	// create BlackFish
    BlackFish = new CBlackFish( RenderTarget, Cache, "graphics\\BlackFish.ini" );
	ASSERT( BlackFish != NULL );

	// ---------------- !! temporary !! -----------------
	// populate map with some test objects
    Marine.push_back( new CMarineSprite(
		BlackFish->GetMap( ),
		769,
        CVector( 128.0f,
				 128.0f,
				 0.0f )));

    Marine.push_back( new CMarineSprite(
		BlackFish->GetMap( ),
		769,
        CVector( 64.0f,
				 64.0f,
				 0.0f )));

    Marine.push_back( new CMarineSprite(
		BlackFish->GetMap( ),
		769,
        CVector( 256.0f,
				 128.0f,
				 0.0f )));

    Marine.push_back( new CMarineSprite(
		BlackFish->GetMap( ),
		769,
        CVector( 64.0f,
				 32.0f,
				 0.0f )));

    Marker = new CSprite(
		BlackFish->GetMap( ),
		0,
		3,
        CVector( 0.0f,
				 0.0f,
				 0.0f ),
		-16 );

    TestBobber = new CBobbingSprite(
		BlackFish->GetMap( ),
		0,
        CVector( 0.0f,
				 0.0f,
				 105.0f ),
		NULL );

	// --------------------------------------------------

	// create chain of event handlers... must be last because they practically control all the other objects
    MainKeyEventHandler = new CMainKeyEventHandler( DirectDraw->GetWindow( ), NULL );
	ASSERT( MainKeyEventHandler != NULL );

    MapEventHandler     = new CMapEventHandler( BlackFish, TestBobber, Marker, MainKeyEventHandler );
	ASSERT( MapEventHandler != NULL );

    RootEventHandler	= MapEventHandler;

	// !! find a better place to initialize this? since delays between this call and Game->Run
	// will break it
	NextFrameTime = Timer.GetTime() + Timer.GetFrequency() / 10;
}

/*
 *
 */
CGame::~CGame( )
{
    for ( std::vector<CSprite *>::size_type i=0; i<Marine.size( ); ++i )
		SAFEDELETE( Marine[i] );
	SAFEDELETE( TestBobber );
	SAFEDELETE( Marker );

	SAFEDELETE( MouseCursor );
    SAFEDELETE( Input );
    SAFEDELETE( Mouse );
	SAFEDELETE( KeyBoard );
    SAFEDELETE( Painter );
    SAFEDELETE( BlackFish );
    SAFEDELETE( BackBuffer );
    SAFEDELETE( Primary );
	SAFEDELETE( DirectDraw );
    SAFEDELETE( Font1 );
    SAFEDELETE( Writer );
    SAFEDELETE( Cache );
	SAFEDELETE( MapEventHandler );
	SAFEDELETE( MainKeyEventHandler );

	RenderTarget	= NULL;
	RootEventHandler= NULL;
}

/*
 * called each frame, the mainloop
 * generates and dispatches the messages to the various handlers and is responsible for
 * polling input devices and for displaying the back buffer
 */
void
CGame::Run( void )
{
	// Input Handling Section
	tSINT32 MouseX, MouseY, MouseWheelPos;
	tUINT32 MouseButtons, LastClick, LastKey;

    Mouse->UpdateState( );
	KeyBoard->UpdateState( );

	Mouse->GetState( MouseX, MouseY, MouseWheelPos, MouseButtons );
	LastClick	= Mouse->GetLastClick( );
	LastKey		= KeyBoard->GetLastKey( );

	// !! I'm passing references to automatic objects to the event handlers - dangerous
	//    thing to do, works under the assumption that event handlers don't keep a pointer...
	// mouseclick message
	if ( LastClick )
	{
        CMouseClickEvent MouseClick( MouseX, MouseY, MouseWheelPos, MouseButtons, LastClick );
		RootEventHandler->HandleEvent( &MouseClick );
	}

	// always generate mousemove messages (should only have few handlers)
    CMouseMoveEvent MouseMove( MouseX, MouseY );
	RootEventHandler->HandleEvent( &MouseMove );

	switch ( LastKey )
	{
		case KEY_ESCAPE:
            CKeyEscEvent KeyEscEvent;
			RootEventHandler->HandleEvent( &KeyEscEvent );
		break;
	}

	// AI section
    for ( std::vector<CSprite *>::size_type i=0; i<Marine.size( ); i++ )
		Marine[i]->Think( );
	TestBobber->Think( );
	Marker->Think();

	// Display section
    Painter->Fill( 0 );		// !! remove
    BlackFish->Render( );

	ShowDebugInfo( );

    RECT Screen;
	SetRect( &Screen, 0, 0, 800, 600 );
    RECT Source = MouseCursor->GetTileRect();
	RECT Dest   = MouseCursor->GetTileRect();
	OffsetRect( &Dest, MouseX - 7, MouseY - 7 );
	if ( Clip( Screen, Source, Dest ))
		MouseCursor->BlitTrans( 0, &Source, *RenderTarget, &Dest );

	if ( DirectDraw->IsFullScreen())
		Primary->Flip( );
	else
	{
        GetClientRect( DirectDraw->GetWindow( ), &Screen);
        ClientToScreen( DirectDraw->GetWindow( ), reinterpret_cast<POINT*>( &Screen.left ));
        ClientToScreen( DirectDraw->GetWindow( ), reinterpret_cast<POINT*>( &Screen.right ));
		Primary->Blit( *BackBuffer, NULL, &Screen );
	}

	// waste cycles so we don't exceed MaxFPS
	while ( Timer.GetTime() <= NextFrameTime );
	NextFrameTime += Timer.GetFrequency() / MaxFPS;
//	while ( Timer.GetTime() - NextFrameTime <= Timer.GetFrequency() / MaxFPS );
//	NextFrameTime = Timer.GetTime();
}

/*
 * restores after a taskswitch
 */
void
CGame::RetrieveFocus( void )
{
    Cache->Invalidate();
*LOG << "Cache Invalidated!" << std::endl;
}
