#include "paint.hpp"

/******************************************************************************
 * BlackFish
 * Copyright (C) 1999-2001 Sren Meyer-Eppler, Jutta Jordans
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ******************************************************************************/


#pragma warning ( push, 1 )
#ifndef __WINDOWS_H__
#define __WINDOWS_H__
  #ifndef WIN32_LEAN_AND_MEAN // !! reset this after including windows.h?
  #define WIN32_LEAN_AND_MEAN
  #endif
#include <windows.h>
#endif
#define DIRECTINPUT_VERSION 0x0700
#include <ddraw.h>
#include <fstream>		// for GetBMPDimension
#pragma warning ( pop )

#include "general\debug.hpp"
#include "general\helpmacros.hpp"
#include "BlackFish\Surface.hpp"

/*
 * initialize, unusable in this state
 */
CPaint::CPaint( void )
	: m_Destination( NULL )
	, m_SurfaceMem( NULL )
	, m_Pitch( 0 )
{
}

/*
 * Set a CSurface object as the drawing target for subsequent
 * calls to drawing functions like PutPixel() and Fill()
 * returns the previous drawing target, so callers may restore that after they are done
 */
CSurface *
CPaint::SetDrawingTarget( CSurface *Dest )
{
    ASSERT( Dest != NULL );              // do we have a valid target?
    ASSERT( !Dest->IsLocked( ));         // don't change drawing target if current target is locked

    CSurface *ReturnValue = m_Destination;
    m_Destination = Dest;
	return ReturnValue;
}

/*
 * Some drawing functions (PutPixel()) require the target surface to be
 * locked and must be called in between a BeginDrawing()/EndDrawing() pair
 */
void
CPaint::BeginDrawing( void )
{
    ASSERT( m_Destination != NULL );

    m_Destination->Lock();

    m_SurfaceMem = m_Destination->GetSurfaceMemPointer();
    m_Pitch = m_Destination->GetPitch();
}

/*
 * unlock the surface
 */
void
CPaint::EndDrawing( void )
{
    ASSERT( m_Destination != NULL );

    m_Destination->UnLock();
}

/*
 * Surface must NOT be locked
 *
 * Clears the whole surface to the specified color
 */
void
CPaint::Fill( tUINT16 Color )
{
    ASSERT( m_Destination != NULL );
    ASSERT( !m_Destination->IsLocked( ));

    HRESULT ReturnValue = DD_OK;    // !! DEBUGCODE this?
    DDBLTFX BlitFX;

    BlitFX.dwSize       = sizeof( BlitFX );
    BlitFX.dwFillColor  = static_cast<DWORD>( Color );

    IDirectDrawSurface7 *Surface;
    Surface = m_Destination->GetDXSurface();
    ASSERT( Surface != NULL );

    m_Destination->RestoreAllIfLost();

    ReturnValue = Surface->Blt( NULL, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &BlitFX );

    ASSERT( ReturnValue == DD_OK );
}

/*
 * Surface must NOT be locked
 *
 * Clears the whole surface to the specified color
 */
void
CPaint::Fill( tSINT32 x1, tSINT32 y1, tSINT32 x2, tSINT32 y2, tUINT16 Color )
{
    ASSERT( m_Destination != NULL );
    ASSERT( !m_Destination->IsLocked( ));

    HRESULT ReturnValue = DD_OK;    // !! DEBUGCODE this?
    DDBLTFX BlitFX;

    BlitFX.dwSize       = sizeof( BlitFX );
    BlitFX.dwFillColor  = static_cast<DWORD>( Color );

    IDirectDrawSurface7 *Surface;
    Surface = m_Destination->GetDXSurface();
    ASSERT( Surface != NULL );

	RECT Rect;
	SetRect( &Rect, x1, y1, x2, y2 );

    m_Destination->RestoreAllIfLost();

    ReturnValue = Surface->Blt( &Rect, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &BlitFX );

    ASSERT( ReturnValue == DD_OK );
}

/*
 * surface must NOT be locked
 *
 * Loads the bitmap file specified by filename into the surface, converting
 * the color format if necessary (for example the 24bit Bitmap to 16bit
 * surface colordepth conversion is done automatically)
 *
 * Note: the surface and bitmap size must be identical
 */
void
CPaint::LoadBmp( const std::string & FileName )
{
    ASSERT( m_Destination != NULL );
    ASSERT( !m_Destination->IsLocked( ));

    IDirectDrawSurface7 *Surface = m_Destination->GetDXSurface();

    HRESULT ReturnValue     = DD_OK;
    HDC     ImageDCHandle   = NULL;
    HDC     SurfaceDCHandle = NULL;
    HBITMAP BitmapHandle    = NULL;
    BITMAP  Bitmap;

    BitmapHandle = static_cast<HBITMAP>( LoadImage( NULL, FileName.c_str(), IMAGE_BITMAP, 0, 0,
                                        LR_LOADFROMFILE | LR_CREATEDIBSECTION ));
	if ( BitmapHandle == NULL )
	{
		*LOG << "Error opening bmp file '" << FileName << "'!" << std::endl;
		QUIT
	}

    ReturnValue = static_cast<HRESULT>( GetObject( BitmapHandle, sizeof( Bitmap ), &Bitmap ));
    ASSERT( ReturnValue != 0 ); // bytes returned

    DDSURFACEDESC2 SurfaceDescriptor;
    ZeroMemory( &SurfaceDescriptor, sizeof( SurfaceDescriptor ) );
    SurfaceDescriptor.dwSize = sizeof( SurfaceDescriptor );
    Surface->GetSurfaceDesc( &SurfaceDescriptor );

    ImageDCHandle = CreateCompatibleDC( NULL );
    SelectObject( ImageDCHandle, BitmapHandle );

    ReturnValue = Surface->GetDC( &SurfaceDCHandle );
    ASSERT( ReturnValue == DD_OK );

    StretchBlt( SurfaceDCHandle,
                0, 0,
				static_cast<tSINT32>( SurfaceDescriptor.dwWidth ),
				static_cast<tSINT32>( SurfaceDescriptor.dwHeight ),
                ImageDCHandle,
                0, 0, Bitmap.bmWidth, Bitmap.bmHeight,
                SRCCOPY );

    Surface->ReleaseDC( SurfaceDCHandle);
    DeleteDC( ImageDCHandle );
    DeleteObject( BitmapHandle );
}

/*
 * surface MUST be locked
 *
 * draws a single pixel of color 'Color' at the specified location
 *
 * Note: No clipping is done, so if the pixel coordinates are illegal,
 * i.e. outside the surface, the result is undefined (=crash)
 */
void
CPaint::PutPixel( tSINT32 X, tSINT32 Y, tUINT16 Color )
{
	ASSERT( m_SurfaceMem != NULL );
    ASSERT( m_Destination != NULL );
    ASSERT( m_Destination->IsLocked( ));

    *(m_SurfaceMem + (Y * m_Pitch) + X) = Color;
}

/*
 * surface MUST be locked
 *
 * no clipping is done, assumes that the coordinates are valid
 * X2 and Y2 are outside the rectangle!!! stupid but windows compatible
 */
void
CPaint::DrawRect( tSINT32 X1, tSINT32 Y1, tSINT32 X2, tSINT32 Y2, tUINT16 Color )
{
	ASSERT( m_SurfaceMem != NULL );
	ASSERT( m_Destination != NULL );
    ASSERT( m_Destination->IsLocked( ));

	tUINT16 *Mem = m_SurfaceMem + (Y1 * m_Pitch) + X1;
	tSINT32 Dist = ( Y2 - Y1 - 1 ) * m_Pitch;

	// top and bottom line of rect
	for ( tSINT32 x = X1; x < X2; x++ )
	{
		*Mem = Color;
		*(Mem + Dist) = Color;
		Mem++;
	}

	Mem  = m_SurfaceMem + (Y1 * m_Pitch) + X1;
	Dist = X2 - X1 - 1;

	// left and right line of rect
	for (tSINT32 y = Y1; y < Y2; y++)
	{
		*Mem = Color;
		*(Mem + Dist) = Color;
		Mem+= m_Pitch;
	}
}

/*
 *	opens .bmp File, reads out Width and Height,
 *	assumes that Width is found at Byte 19 - 22
 *  Heigth at 23 - 26
 *  returns true if successful
 */
bool
CPaint::GetBmpDimensions( 
	const char *FileName,
	tSINT32 *Width,
	tSINT32 *Height)
{
	// Variables for the reading out of Width and Heigth,
	// probably not neccessary, could use original Width and Height,
	// but seems safer that way
	tSINT32 ReadWidth, ReadHeight;

	std::ifstream BitmapFile;
	BitmapFile.open( FileName );

	if (BitmapFile.good())
	{
		//set getpointer to pos 19 and get width
		BitmapFile.seekg(18);
		BitmapFile.read( (char*) &ReadWidth, sizeof(ReadWidth) );

		//set getpointer to pos 23 (should already be there, but anyway) and get height
		BitmapFile.seekg(22);
		BitmapFile.read( (char*) &ReadHeight, sizeof(ReadHeight) );

		*Width = ReadWidth;
		*Height = ReadHeight;

		BitmapFile.close();
        return true;
	}
	else
	{
		BitmapFile.close();
        return false;
	}
}
