#include "sprite.hpp"

/******************************************************************************
 * BlackFish
 * Copyright (C) 1999-2001 Sren Meyer-Eppler, Jutta Jordans
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ******************************************************************************/


#include "BlackFish\time.hpp"


/**
 * constructor, registers the sprite with the appropriate map cell
 */
CSprite::CSprite(
    CMap *Map,
	tSINT32 TileSet,
	tSINT32 TileNr,
    const CVector & Position,
	tSINT32 DrawingOrder )
	: TileSet( TileSet )
	, TileNr( TileNr )
	, Map( Map )
	, Position( Position )
	, Flags( 0 )
	, DrawingOrder( DrawingOrder )
{
	ASSERT( Map != NULL );
	Map->InsertSprite( &this->Position, this );
}


CSprite::~CSprite( )
{
	Map->RemoveSprite( &this->Position, this );
	Map = NULL;
}


/**
 * return the sprite's tileset
 */
tSINT32
CSprite::GetTileSet( )
{
	return TileSet;
}


/**
 * returns the current frame's tile nr, base class always returns 0 since it's
 * not animated
 */
tSINT32
CSprite::GetTileNr( )
{
	return TileNr; // base class type is not animated
}


/**
 * change the sprite's position, registers the new position with the associated map
 */
bool
CSprite::SetPosition( const CVector & Position )
{
    bool Moved = false;

	tSINT32 NewX = static_cast<tSINT32>( Position.x / 32.0 );
	tSINT32 NewY = static_cast<tSINT32>( Position.y / 32.0 );

	if ( Map->IsValidCell( NewX, NewY )
		&& (Position.x >= 0.0) && (Position.y >= 0.0))	// !! find a better way, only necessary
	{													// because negative numbers are truncated to 0
        Moved = true;
		if (( NewX != static_cast<tSINT32>(this->Position.x / 32.0f ))
		  ||( NewY != static_cast<tSINT32>(this->Position.y / 32.0f )))
		{
			Map->RemoveSprite( &(this->Position), this );
			this->Position.Set( Position );
			Map->InsertSprite( &(this->Position), this );
		}
		else
		{
			this->Position.Set( Position );
		}
	}
	return Moved;
}


/**
 * return sprite position in world-coordinates
 */
void
CSprite::GetPosition( CVector & Position )
{
	Position.Set( this->Position );
}


/**
 * perform AI, base class does nothing (override)
 */
void
CSprite::Think( )
{
}


tUINT32 
CSprite::GetFlags( void ) 
{ 
	return Flags; 
}


void	
CSprite::SetDrawingOrder( tSINT32 Priority )
{
	DrawingOrder = Priority;
}


tSINT32 
CSprite::GetDrawingOrder() const
{
	return DrawingOrder;
}


// used to sort sprites according to drawingorder
bool 
SpriteComparer( const CSprite * lhs, const CSprite * rhs )
{
	return ( lhs->Position.z   < rhs->Position.z ) ||
		   ( lhs->Position.x   > rhs->Position.x ) ||
		   ( lhs->Position.y   < rhs->Position.y ) ||
		   ( lhs->DrawingOrder < rhs->DrawingOrder );
}

/**********************************************************************************/

/*
 *
 */
CMarineSprite::CMarineSprite( CMap *Map, tSINT32 TileSet, const CVector & Position, tSINT32 DrawingOrder )
    : CSprite( Map, TileSet, 0, Position, DrawingOrder ),
      CBroadcaster( ),
	  LastAnimTime( Timer.GetTime( )),
	  LastFrameTime( Timer.GetTime( )),
	  FrameTime( Timer.GetTime( )),
	  TileNrBase( 64 ),
	  SequenceLength( 1 ),
	  Direction( EAST ),
	  Speed( 0.0f, 0.0f, 0.0f ) ,
	  Path( )
{
	Flags |= FLAGBLOCKING;
}

/*
 *
 */
void
CMarineSprite::Move( const CVector & To )
{
	if ( Path.empty( )) // don't accept a new path while still on one
		// we're not interested whether a path could actually be found at this point
		Map->FindPath( Position, To, &Path );
}

/*
 * !! this function will get long -> split
 * !! do something against those sharp turn animations, i.e. a turn from N to SE should
 *    display the NE, E animations in between
 * This function also updates the internal frametimer variables, since it is called exactly once
 * per frame
 */
void
CMarineSprite::Think( )
{
	FrameTime = Timer.GetTime( );

	if ( !Path.empty( ))		// we haven't arrived, yet
	{
		if ( !Position.Equals( Path.front( ))) // we are between two map cells
		{
			Speed.Normalize( );		// !! don't renormalize each frame (sqrt()!) but save the unit
									//    vector add a multiply of that to Position (->modify vector class)
			Speed.Multiply( static_cast<tFLOAT>(( FrameTime - LastFrameTime ) * UnitsPerSecond )
						  / static_cast<tFLOAT>( Timer.GetFrequency()));

            CVector NewPos;
			NewPos.Set( Position );

            CVector Distance;
			Distance.Set( Path.front( ));  // instead of calculating Position from scratch each frame
			Distance.Subtract( Position ); // I could save it and just subtract Speed each frame

			if ( Speed.GetLength( ) > Distance.GetLength( ))	// make sure we don't overshoot our goal
				NewPos.Add( Distance );
			else
				NewPos.Add( Speed );

            bool ReturnValue = SetPosition( NewPos );
			ASSERT( ReturnValue );
		}
		else	// we are at the goal position
		{
            CVector NewPos;
			NewPos.Set( Path.front( ));
            bool ReturnValue = SetPosition( NewPos );   // set our position to the goal position (to prevent
			ASSERT( ReturnValue );						// floating point errors from accumulating)
			Path.pop_front( );
			if ( !Path.empty( ))	// steer towards our new goal if any
			{
				Speed.Set( Path.front( ));
				Speed.Subtract( Position );

				if ( Speed.x < 0.0 )
					if ( Speed.y < 0.0 )
						Direction = WEST;
					else if ( Speed.y > 0.0 )
						Direction = SOUTH;
					else
						Direction = SOUTHWEST;
				else if ( Speed.x > 0.0 )
					if ( Speed.y < 0.0 )
						Direction = NORTH;
					else if ( Speed.y > 0.0 )
						Direction = EAST;
					else
						Direction = NORTHEAST;
				else
					if ( Speed.y < 0.0 )
						Direction = NORTHWEST;
					else
						Direction = SOUTHEAST;

				SequenceLength  = 8;	// 8 walk frames in each direction
				TileNrBase		= (Direction - 1) * SequenceLength;
			}
		}
		Notify( ); // Tell Observers that our position has changed
	}
	else // no goal, use stop and stand animations...
	{
		TileNr = 0;
		SequenceLength = 1;
		TileNrBase = 63 + Direction;
	}	
	LastFrameTime = FrameTime;
}

/*
 *
 */
tSINT32
CMarineSprite::GetTileNr( )
{
	tDOUBLE AnimDelta = static_cast<double>( FrameTime - LastAnimTime ) 
					  / static_cast<double>( Timer.GetFrequency());
	if ( AnimDelta >= 1.0 / Frequency )
	{
		TileNr += static_cast<tSINT32>( AnimDelta * Frequency );
		TileNr %= SequenceLength;
		LastAnimTime = FrameTime;
	}

	return TileNr + TileNrBase;
}

/**********************************************************************************/
/**
 * constructor, registers the sprite with the appropriate map cell
 */
CBobbingSprite::CBobbingSprite(
    CMap *Map,
	tSINT32 TileSet,
    const CVector & Position,
    CMarineSprite *Subject,
	tSINT32 DrawingOrder )
    : CSprite( Map, TileSet, 0, Position, DrawingOrder ),
      CListener( ),
	  Marine( Subject ),
      Visible( true ),
	  LastFrameTime( Timer.GetTime( )),
	  FrameTime( Timer.GetTime( )),
	  Speed( 0.0f, 0.0f, 1.0f )
{
	if ( Marine == NULL )
        SetVisible( false );
}

/**
 * returns the current frame's tile nr
 */
tSINT32
CBobbingSprite::GetTileNr( )
{
	return 0;
}

/**
 *
 */
void
CBobbingSprite::SetSubject( CMarineSprite *Subject )
{
	// may be NULL
	Marine = Subject;
	if ( Marine == NULL )
	{
        SetVisible( false );
	}
	else
	{
        SetVisible( true );
		Update( Subject );
	}
}

/**
 *
 */
void
CBobbingSprite::SetVisible( bool Visible )
{
	if ( Visible != this->Visible )
	{
		if ( Visible )
			Map->InsertSprite( &this->Position, this );
		else
			Map->RemoveSprite( &this->Position, this );

		this->Visible = Visible;
	}
}

/**
 *
 */
void
CBobbingSprite::Think( )
{
	if ( !Visible )
		return;

	FrameTime = Timer.GetTime( );

	Speed.Normalize( );
	Speed.Multiply( static_cast<tFLOAT>(( FrameTime - LastFrameTime ) * UnitsPerSecond ) 
				  / static_cast<tFLOAT>(Timer.GetFrequency()));
	Position.Add( Speed );

	if ( Position.z >= ZMAX )
	{
		Position.z = ZMAX;
		Speed.z *= -1.0f;
	}
	else if ( Position.z <= ZMIN )
	{
		Position.z = ZMIN;
		Speed.z *= -1.0f;
	}

	LastFrameTime = FrameTime;
}

/**
 *
 */
void
CBobbingSprite::Update( CBroadcaster *Subject )
{
	ASSERT( Subject != NULL );
	ASSERT( Marine != NULL );
	ASSERT( Visible );
	// wrong Subject, we don't monitor it
    if ( dynamic_cast< CMarineSprite *>( Subject ) != Marine )
		return;

    CVector NewPosition;
	Marine->GetPosition( NewPosition );
	NewPosition.z = Position.z;

    bool Result = SetPosition( NewPosition );
	ASSERT( Result );
}
