#include "Surface.hpp"

/******************************************************************************
 * BlackFish
 * Copyright (C) 1999-2001 Sren Meyer-Eppler, Jutta Jordans
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ******************************************************************************/


#pragma warning ( push, 1 )
#define DIRECTINPUT_VERSION 0x0700
#include <ddraw.h>
#include <cassert>
#pragma warning ( pop )

#include "General/HelpMacros.hpp"
#include "BlackFish/DxDraw.hpp"


//-------------------------------------------------------------------------------------------------


/**
 * Constructor, initialisiert alle Variablen auf ihre Default Werte,
 * die class ist in diesem Zustand nicht zu gebrauchen, 'Create' mu
 * aufgerufen werden
 */
CSurface::CSurface()
	: m_Surface( NULL )                   // surface doesn't exist, yet - call create()
     ,m_SurfaceMem( NULL )                // } call lock() to set these values
     ,m_Pitch( 0 )                        // }
     ,m_ImCachedHere( NULL )              // surface is not yet cached
     ,m_Width( 0 ) 
     ,m_Height( 0 )
     ,m_Locked( false )
{
}


/**
 * Destructor, releases all allocated memory, i.e. the Surface
 */
CSurface::~CSurface()
{
	assert( !m_Locked ); // don't destroy locked surfaces
	assert( m_Surface != NULL );
	
	// hopefully kills any attached clippers, might already be done automatically by surface dtor
	m_Surface->SetClipper( NULL );
    RELEASE( m_Surface );
}


/**
 * Erstellt die primary Surface mit einem Backbuffer. Die Gre und Farbtiefe
 * der Surface entspricht der aktuellen Bildschirmeinstellung
 * - DirectDraw mu ein gltiges DirectDraw Objekt sein
 * - die Auflsung mu bereits mit 'DirectDraw->SetMode' gesetzt worden sein
 * - es sollte nur eine Primary Surface im gesamten Programm geben
 *      !! use a design pattern for that requirement?
 *      !! use a design pattern that the user can only call one Create function
 *         once for every instance of this class
 * - die backbuffer-surface erhlt man, indem man in einem zweiten CSurface
 *   Objekt 'Create( PrimarySurface )' aufruft
 * - !! untested: both surfaces are created in video-memory if possible, else ???
 */
void
CSurface::Create()
{
    assert( DirectDraw != NULL );
    assert( m_Surface == NULL );  // make sure surface hasn't been created, yet

    DDSURFACEDESC2 SurfaceDescriptor;
    ZeroMemory( &SurfaceDescriptor, sizeof( SurfaceDescriptor ) );
    SurfaceDescriptor.dwSize = sizeof( SurfaceDescriptor );

    HRESULT ReturnValue = DD_OK;
	if ( DirectDraw->IsFullScreen( ))
	{
		SurfaceDescriptor.dwFlags           = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
		SurfaceDescriptor.ddsCaps.dwCaps    = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
		SurfaceDescriptor.dwBackBufferCount = 1;

		ReturnValue = DirectDraw->GetDirectDrawInterface()->CreateSurface( 
			&SurfaceDescriptor, &m_Surface, NULL);
		assert( ReturnValue == DD_OK );
	}
	else
	{
		SurfaceDescriptor.dwFlags           = DDSD_CAPS;
		SurfaceDescriptor.ddsCaps.dwCaps    = DDSCAPS_PRIMARYSURFACE;

		ReturnValue = DirectDraw->GetDirectDrawInterface()->CreateSurface( 
			&SurfaceDescriptor, &m_Surface, NULL);
		assert( ReturnValue == DD_OK );

		IDirectDrawClipper *Clipper = NULL;
        ReturnValue = DirectDraw->GetDirectDrawInterface()->CreateClipper( 0, &Clipper, NULL );
		assert( ReturnValue == DD_OK );

        Clipper->SetHWnd( 0, DirectDraw->GetWindow( )); // !! what are the return values?!
        m_Surface->SetClipper( Clipper );
        tUINT32 ReferenceCount = Clipper->Release( );
		assert( ReferenceCount == 1 ); // the clipper is referenced by Surface
	}

    ZeroMemory( &SurfaceDescriptor, sizeof( SurfaceDescriptor ) );
    SurfaceDescriptor.dwSize = sizeof( SurfaceDescriptor );
	ReturnValue = m_Surface->GetSurfaceDesc( &SurfaceDescriptor );
	assert( ReturnValue == DD_OK );

	SetTransparentColor( 0 );	// default transparent color, so blittrans calls don't crash

	m_Width	 = static_cast<tSINT32>( SurfaceDescriptor.dwWidth );
	m_Height = static_cast<tSINT32>( SurfaceDescriptor.dwHeight );
}


/**
 * Erstellt einen pointer auf die Backbuffer surface zu der angegebenen Primary Surface
 * - Die Primary mu gltig sein, i.e. 'Primary->Create( DirectDraw )'
 *   mu vorher aufgerufen worden sein
 */
void
CSurface::Create( CSurface *Primary )
{
    assert( Primary != NULL );
    assert( m_Surface == NULL );

    HRESULT ReturnValue = DD_OK;

    DDSCAPS2 SurfaceCaps;
    ZeroMemory( &SurfaceCaps, sizeof( SurfaceCaps ));

    SurfaceCaps.dwCaps = DDSCAPS_BACKBUFFER;

    ReturnValue = Primary->GetDXSurface()->GetAttachedSurface( &SurfaceCaps, &m_Surface );
    assert( ReturnValue == DD_OK );

	SetTransparentColor( 0 );	// default transparent color, so blittrans calls don't crash

	m_Width	 = Primary->GetWidth( );
	m_Height = Primary->GetHeight( );
}


/**
 * Erstellt eine Surface der angegebenen Ausmae im Systemspeicher
 * !! ist die Farbtiefe wirklich immer = der DirectDraw Tiefe?
 */
void
CSurface::Create( tSINT32 Width, tSINT32 Height, bool VideoMemory )
{
    assert( DirectDraw != NULL );
    assert( m_Surface == NULL );  // make sure surface hasn't been created, yet

    m_Width   = Width;
    m_Height  = Height;

    DDSURFACEDESC2 SurfaceDescriptor;
    ZeroMemory( &SurfaceDescriptor, sizeof( SurfaceDescriptor ) );
    SurfaceDescriptor.dwSize            = sizeof( SurfaceDescriptor );
    SurfaceDescriptor.dwWidth           = static_cast<DWORD>( GetWidth());
    SurfaceDescriptor.dwHeight          = static_cast<DWORD>( GetHeight());
    SurfaceDescriptor.dwFlags           = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	SurfaceDescriptor.ddsCaps.dwCaps	= DDSCAPS_OFFSCREENPLAIN;
    SurfaceDescriptor.ddsCaps.dwCaps   |= VideoMemory ? DDSCAPS_VIDEOMEMORY : DDSCAPS_SYSTEMMEMORY;

    HRESULT ReturnValue = DirectDraw->GetDirectDrawInterface()->CreateSurface( 
		&SurfaceDescriptor, &m_Surface, NULL);
    assert( ReturnValue == DD_OK );

	SetTransparentColor( 0 );	// default transparent color, so blittrans calls don't crash
}


/**
 * checks whether this surface has been lost (overwritten in videomem, sysmem surfaces can't
 * get lost) and calls DirectDraw->RestoreAllSurfaces( ), which restores not only this, but
 * all directdraw surfaces
 */
void
CSurface::RestoreAllIfLost() const
{
    assert( m_Surface != NULL );

    if ( m_Surface->IsLost() == DDERR_SURFACELOST )
    {
        IDirectDraw4 *DirectDraw4;
        m_Surface->GetDDInterface( reinterpret_cast<void **>( &DirectDraw4 ));
        DirectDraw4->RestoreAllSurfaces();
    }
}


/**
 * copy the Region specified by 'SourceRect' from the surface 'Source' onto the
 * calling (this) surface in the Region specified by 'DestRect'
 * - if SourceRect and DestRect are of different sizes, stretching/shrinking is automatically
 *   performed
 * - if SourceRect or DestRect are partially or fully outside of surface bounds, the result
 *   is undefined, i.e. clipping is *not* performed
 * - if either dest or sourcerect is NULL, the whole surface will be blitted to/from.
 */
void
CSurface::Blit( const CSurface &Source, const RECT *SourceRect, const RECT *DestRect ) const
{
    RestoreAllIfLost();

    HRESULT ReturnValue( DDERR_WASSTILLDRAWING );
    while ( ReturnValue == DDERR_WASSTILLDRAWING )
    {
        ReturnValue = m_Surface->Blt( 
			const_cast<RECT *>( DestRect ),	// yuck, but DX API isn't good with const correctness
			Source.GetDXSurface(), 
			const_cast<RECT *>( SourceRect ),
			DDBLT_ASYNC, NULL );
    }
    assert( ReturnValue == DD_OK );
}


/**
 * same as 'Blit()', but uses the SrcColorKey, meaning that the specified color is not
 * copied from the source surface, i.e. appears transparent
 */
void
CSurface::BlitTrans( const CSurface &Source, const RECT *SourceRect, const RECT *DestRect ) const
{
    RestoreAllIfLost();

    HRESULT ReturnValue( DDERR_WASSTILLDRAWING );
    while ( ReturnValue == DDERR_WASSTILLDRAWING )
    {
        ReturnValue = m_Surface->Blt( 
			const_cast<RECT *>( DestRect ),	// yuck, but DX API isn't good with const correctness
			Source.GetDXSurface(), 
			const_cast<RECT *>( SourceRect ),
			DDBLT_ASYNC | DDBLT_KEYSRC, NULL );
    }
    assert( ReturnValue == DD_OK );
}


/**
 * Flips the Primary surface and the Backbuffer
 * - this function must only be called on the Primary Surface, i.e.
 *   on the surface created with 'Create( DirectDraw )', else
 *   strange things will happen
 */
void
CSurface::Flip()
{
    assert( m_Surface != NULL );

    RestoreAllIfLost();

    HRESULT ReturnValue = DDERR_WASSTILLDRAWING;
    while ( ReturnValue == DDERR_WASSTILLDRAWING )
        ReturnValue = m_Surface->Flip( NULL, DDFLIP_WAIT );

    assert( ReturnValue == DD_OK ); // if it isn't DD_OK we couldn't flip
}


/**
 * specifies the transparent color (source-colorkey)
 * note: pixel format is important here, so if the transparent color is neither
 *       black or white you should use a CPixelConverter to calculate the value
 */
void
CSurface::SetTransparentColor( tUINT32 Color )
{
	assert( m_Surface != NULL );

    DDCOLORKEY ColorKey;
    ColorKey.dwColorSpaceLowValue   = static_cast<DWORD>( Color );
    ColorKey.dwColorSpaceHighValue  = ColorKey.dwColorSpaceLowValue;

    HRESULT ReturnValue = m_Surface->SetColorKey( DDCKEY_SRCBLT, &ColorKey );
    assert( ReturnValue == DD_OK );
}


/**
 * returns the transparent color (source-colorkey)
 */
tUINT32
CSurface::GetTransparentColor() const
{
	assert( m_Surface != NULL );

    DDCOLORKEY ColorKey;
    ZeroMemory( &ColorKey, sizeof ( ColorKey ));

    HRESULT ReturnValue = m_Surface->GetColorKey( DDCKEY_SRCBLT, &ColorKey );
    assert( ReturnValue == DD_OK );
    assert( ColorKey.dwColorSpaceLowValue == ColorKey.dwColorSpaceHighValue );

    return static_cast<tUINT32>( ColorKey.dwColorSpaceLowValue );
}


RECT
CSurface::GetSize() const
{
	RECT Temp;
	SetRect( &Temp, 0, 0, GetWidth(), GetHeight());
	return Temp;
}


/**
 * Locks the surface's memory for direct access afterwards
 * - must only be called on systemmory surfaces, i.e. not the primary or backbuffer
 * - you must call 'UnLock()' after locking a surface
 * - only use functions in between a 'Lock()' 'UnLock()' pair that don't attempt
 *   to lock the surface memory themselves
 */
void
CSurface::Lock()
{
    assert( m_Surface != NULL );
	assert( !m_Locked );

    m_Locked = true;

    DDSURFACEDESC2 SurfaceDescriptor;
    ZeroMemory( &SurfaceDescriptor, sizeof( SurfaceDescriptor ) );
    SurfaceDescriptor.dwSize = sizeof ( SurfaceDescriptor );

    HRESULT ReturnValue = DDERR_WASSTILLDRAWING;
    while ( ReturnValue == DDERR_WASSTILLDRAWING )
        ReturnValue = m_Surface->Lock( NULL, &SurfaceDescriptor, 0, NULL );
    assert( ReturnValue == DD_OK );

    m_SurfaceMem = static_cast<tUINT16 *>( SurfaceDescriptor.lpSurface );
    m_Pitch		 = static_cast<tSINT32>( SurfaceDescriptor.lPitch >> 1 );
    assert( m_SurfaceMem != NULL );
    assert( m_Pitch != 0 );
}


/**
 * Unlocks the surface memory
 * - 'Lock()' must have been called before
 */
void
CSurface::UnLock()
{
    assert( m_Surface != NULL );
	assert( m_Locked );

    m_Locked = false;

    HRESULT ReturnValue = m_Surface->Unlock( NULL );
    assert( ReturnValue == DD_OK );

    m_SurfaceMem = NULL;
    m_Pitch		 = 0;
}


//-------------------------------------------------------------------------------------------------
/**
 * adjust the Source rect so it fits into the ClipRect, return the result in Dest
 * (Source is unaltered)
 * return true if at least part of the rect is still visible
 */
bool 
Clip( const RECT & ClipRect, RECT & Source, RECT & Dest )
{
    // trivially reject if tile is completely out of the visible rect
	if (( Dest.left >= ClipRect.right ) || ( Dest.right <= ClipRect.left ) ||
		( Dest.top >= ClipRect.bottom ) || ( Dest.bottom <= ClipRect.top ))
        return false;

	tSINT32 Difference       = 0;
	if (( Difference = Dest.left - ClipRect.left) < 0 )   // clip left
	{
		Dest.left       = ClipRect.left;
		Source.left    -= Difference;
	}
	else if (( Difference = Dest.right - ClipRect.right ) > 0 )   // clip right
	{
		Dest.right      = ClipRect.right;
		Source.right   -= Difference;
	}
	if (( Difference = Dest.top - ClipRect.top ) < 0 )   // clip top
	{
		Dest.top        = ClipRect.top;
		Source.top     -= Difference;
	}
	else if (( Difference = Dest.bottom - ClipRect.bottom ) > 0 )   // clip bottom
	{
		Dest.bottom     = ClipRect.bottom;
		Source.bottom  -= Difference;
	}
    return true;
}
