#include "tileset.hpp"

/******************************************************************************
 * BlackFish
 * Copyright (C) 1999-2001 Sren Meyer-Eppler, Jutta Jordans
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ******************************************************************************/

#pragma warning ( push, 1 )
#include <cassert>
#pragma warning ( pop )

#include "General/HelpMacros.hpp"
#include "IniFile/IniFile.hpp"
#include "BlackFish/Paint.hpp"
#include "BlackFish/PixelConverter.hpp"


/**
 * constructor, creates a tileset according to the specifications given in
 * the 'IniFileName' inifile section 'ResourceName'
 * does not initialize a cache (so default behaviour is uncached blits)
 * note: all tiles in a tileset are assumed to be of equal size
 */
CTileSet::CTileSet( const char *IniFileName, const char *ResourceName )
	: m_TileWidth( 0 )
     ,m_TileHeight( 0 )
     ,m_SurfaceWidth( 0 )
     ,m_SurfaceHeight( 0 )
     ,m_NrOfTiles( 0 )
     ,m_Tile( NULL )
     ,m_Cache( NULL )
{
    CIniFile *IniFile = NULL;

    IniFile = new CIniFile( IniFileName );
    assert( IniFile != NULL );

    bool ReturnValue = false;
    char BitmapFileName[80] = "";

    ReturnValue = IniFile->ReadString( ResourceName, "FileName", BitmapFileName ); // read corresponding bitmap filename
    assert( ReturnValue );

    ReturnValue = IniFile->ReadNumber( ResourceName, "TileWidth", &m_TileWidth );
    assert( ReturnValue );

    ReturnValue = IniFile->ReadNumber( ResourceName, "TileHeight", &m_TileHeight );
    assert( ReturnValue );

    ReturnValue = IniFile->ReadNumber( ResourceName, "BitmapWidth", &m_SurfaceWidth );
    assert( ReturnValue );

    ReturnValue = IniFile->ReadNumber( ResourceName, "BitmapHeight", &m_SurfaceHeight );
    assert( ReturnValue );

    ReturnValue = IniFile->ReadNumber( ResourceName, "NrOfTiles", &m_NrOfTiles );
    assert( ReturnValue );

    CSurface Bitmap;
    Bitmap.Create( m_SurfaceWidth, m_SurfaceHeight, false );

    CPixelConverter PixelConverter( &Bitmap );
	tUINT32 TransparentColor = PixelConverter.RGBToColor( 255, 0, 255 );
    Bitmap.SetTransparentColor( TransparentColor );	// !! hardcoded transparent color

    CPaint Painter;
    Painter.SetDrawingTarget( &Bitmap );
    Painter.LoadBmp( BitmapFileName );

    m_Tile = NULL;
    m_Tile = new CSurface*[static_cast<tUINT32>( m_NrOfTiles )];
    assert( m_Tile != NULL );

    tSINT32 TilesPerRow = m_SurfaceWidth / m_TileWidth;

    for ( tSINT32 TileNr = 0; TileNr < m_NrOfTiles; ++TileNr )
    {
        tSINT32 RowNr = TileNr / TilesPerRow;

	    RECT SourceRect;
        SourceRect.top      = RowNr * m_TileHeight;
        SourceRect.bottom   = SourceRect.top + m_TileHeight;
        SourceRect.left     = ( TileNr - ( TilesPerRow * RowNr )) * m_TileWidth;
        SourceRect.right    = SourceRect.left + m_TileWidth;

        m_Tile[TileNr] = new CSurface();
        m_Tile[TileNr]->Create( m_TileWidth, m_TileHeight, false );
        m_Tile[TileNr]->SetTransparentColor( TransparentColor );
        m_Tile[TileNr]->Blit( Bitmap, &SourceRect, NULL );
    }

    SAFEDELETE( IniFile );
}


CTileSet::~CTileSet()
{
    for ( tSINT32 TileNr=0; TileNr<m_NrOfTiles; TileNr++ )
        SAFEDELETE( m_Tile[TileNr] );
    SAFEDELETEARRAY( m_Tile );
	SetCache( NULL );
}


/**
 * returns the CSurface object corresponding to the given TileNr
 */
CSurface *
CTileSet::GetTile( tSINT32 TileNr ) const
{
    return m_Tile[TileNr];
}


/**
 * note: all tiles in a tileset are assumed to be of equal size
 */
RECT
CTileSet::GetTileRect() const
{
	RECT Temp;
	SetRect( &Temp, 0, 0, GetTileWidth(), GetTileHeight());
	return Temp;
}


/**
 * Blits tile 'TileNr' onto the Surface 'Dest' using the regions specified by source and
 * destrect, scaling if neccessary.
 * uses a cache if one specified by 'SetCache' otherwise blits the surface directly
 * !! confusing, since blit in this case means blitting the calling object
 *    somewhere, whereas in CSurface blit means blitting something on
 *    the calling object
 */
void
CTileSet::Blit( tSINT32 TileNr, const RECT *SourceRect, CSurface &Dest, const RECT *DestRect ) const
{
	if ( m_Cache != NULL )
        m_Cache->Blit( *m_Tile[TileNr], SourceRect, Dest, DestRect );
	else
		Dest.Blit( *m_Tile[TileNr], SourceRect, DestRect );
}


/**
 * Blits tile 'TileNr' onto the Surface 'Dest' using the regions specified by source and
 * destrect, scaling if neccessary.
 * uses a cache if one specified by 'SetCache' otherwise blits the surface directly
 * uses the transparent color specified in the constructor (255,0,255)
 * !! confusing, since blit in this case means blitting the calling object
 *    somewhere, whereas in CSurface blit means blitting something on
 *    the calling object
 */
void
CTileSet::BlitTrans( tSINT32 TileNr, const RECT *SourceRect, CSurface &Dest, const RECT *DestRect ) const
{
	if ( m_Cache != NULL )
		m_Cache->BlitTrans( *m_Tile[TileNr], SourceRect, Dest, DestRect );
	else
		Dest.BlitTrans( *m_Tile[TileNr], SourceRect, DestRect );
}
