#include "IniFile.hpp"

/******************************************************************************
 * BlackFish
 * Copyright (C) 1999-2001 Sren Meyer-Eppler, Jutta Jordans
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ******************************************************************************/


#pragma warning ( push, 1 )
#include <stdio.h>      // sscanf, eigentlich wollt ich's ja vermeiden...
#pragma warning ( pop )

// !! are the external include guards needed in this case, too?
#include "general\debug.hpp"
#include "general\helpmacros.hpp"

/*
 * constructor, initializes a CIniFile object to it's default values, the object
 * is unusable in this state, Open() has to be called to specify an input file first
 */
CIniFile::CIniFile( void )
	: InputFile( NULL ),
	  InputFileName( NULL )
{
    Initialize( );
}

/*
 * constructor, initializes a CIniFile object by calling Open() with the specified
 * Filename, the object is fully functional after this
 */
CIniFile::CIniFile( const char *FileName )
	: InputFile( NULL ),
	  InputFileName( NULL )
{
    // reset everything to it's default values
    Initialize( );

    // call open() to prepare a file for processing
    Open( FileName );
}

/*
 * Destructor
 */
CIniFile::~CIniFile( void )
{
    Close();
	SAFEDELETEARRAY( InputFileName );
}

/*
 * Resets all variables to their default values, the object is unsusable in
 * this state, 'Open()' has to be called
 */
void
CIniFile::Initialize( void )
{
    // Reset the filehandle to no file
    InputFile = NULL;
	SAFEDELETEARRAY( InputFileName );

	// fill the currently loaded string with spaces - there's nothing loaded, yet
    memset( CurrentLine, ' ', MAXLINELENGTH );
	CurrentLine[MAXLINELENGTH - 1] = '\0';

    // fill the currently loaded valuestring (the characters following the equals
    // sign after the keyword in the *.ini file) with spaces
    memset( ValueString, ' ', MAXLINELENGTH );
	ValueString[MAXLINELENGTH - 1] = '\0';
}

/*
 * Opens a file for input, if any error occurs it immediately quits the program with
 * an assert() error message
 */
void
CIniFile::Open( const char *FileName )
{
    ASSERT( FileName != NULL );
	InputFileName = new char[strlen( FileName ) + 1];
	memcpy( InputFileName, FileName, ( strlen( FileName ) + 1 ) * sizeof( char ));

    // open the file 'FileName' for processing, read only text-mode
    InputFile = new std::ifstream( FileName );

	if ( !InputFile->is_open( ))
	{
		*LOG << "Error opening ini file: '" << InputFileName << "'!" << std::endl;
		QUIT
	}

    // check for errors, (check if the filehandle is valid)
    ASSERT( InputFile != NULL );

    // important to check == false and not != true cause it might return values != 1
   // ASSERT( InputFile->fail() == false );
}

/*
 * Closes the input file
 */
void
CIniFile::Close( void )
{
    //  !! there doesn't seem to be a special close() function -> is that correct? (destructor?)

	SAFEDELETEARRAY( InputFileName );

    // free the file object
    SAFEDELETE ( InputFile );
}

/*
 * Alle nachfolgenden Leseoperationen starten am Dateianfang
 */
void
CIniFile::Rewind( void )
{
    ASSERT( InputFile != NULL );

    // setzt den FileIndex zurck auf den Anfang der Datei
    InputFile->seekg( 0, std::ios_base::beg );

    //ASSERT( InputFile->fail() == false );
}

/*
 * Read the next Line from the File into the buffer 'CurrentLine'
 */
bool
CIniFile::ReadLine( void )
{
    ASSERT( InputFile != NULL );

    InputFile->getline( CurrentLine, MAXLINELENGTH, '\n' );

    // if end of file (EOF) was reached return false, else true
    // note the strange way to check for this, read the 'getline()' help for details
    if ( !InputFile->fail())
    {   // nominal case, EOF not reached
        return true;
    }
    else
    {   // EOF encountered
        return false;
    }
    // this point is never reached
}

/*
 * Checks if the first characters in 'CurrentLine' match the specified 'StringName'
 *
 * Note: none of the scanf functions is used here cause they would ignore comments
 * in the *.ini file
 */
bool
CIniFile::FindString( const char *StringName )
{
    ASSERT( StringName != NULL );

    bool MatchFound = false;        // does the first word in CurrentLine match StringName?
    const char *StringPtr = StringName;   // Pointers used to increment through the strings
    const char *LinePtr   = CurrentLine;

    // Solange das gesuchte Wort noch nicht gefunden wurde und die beiden
    // aktuellen Vergleichsbuchstaben bereinstimmen gehe weiter durch
    // den string
    while ( !MatchFound && ( *StringPtr == *LinePtr ))
    {
        // advance one character in both strings
        StringPtr++;
        LinePtr++;

        // wenn der NullCharakter ('\0') erreicht ist, ist das Suchwort zu
        // Ende. Da alle Zeichen bis zu diesem Punkt bereits bereinstimmen
        // knnen wir davon ausgehen, da das Wort gefunden wurde
        if ( *StringPtr == '\0' )
            MatchFound = true;
    }

    return MatchFound;
}

/*
 * Reads through the file line by line and tries to find the keyword 'SectionName'
 * returns true if the keyword was found
 */
bool
CIniFile::FindSection( const char *SectionName )
{
    ASSERT( SectionName != NULL );

    bool SectionFound = false;

	do
    {
        SectionFound = FindString( SectionName );
    } // continue while there is still more to read and we haven't found the Section yet
    while ((ReadLine()) && (!SectionFound));

    return SectionFound;
}

/*
 * Reads through the file line by line until the keyword is found, or the
 * beginning of a new Section is found. New Sections are marked with "[" because
 * the section name is always enclosed in "[]"
 * returns true if the keyword was found in this Section
 * Note: This function assumes that the file pointer already points to the be-
 * ginning of the right Section, i.e. 'FindSection()' has been called
 */
bool
CIniFile::FindKeyWord( const char *KeyWord )
{
    ASSERT( KeyWord != NULL );

    bool KeyWordFound = false;
    bool SectionEnd   = false;
    bool ReadOk       = true;

    do
    {
        KeyWordFound = FindString( KeyWord );

        SectionEnd = FindString("[");

        // nur die nchste Zeile lesen wenn das 'KeyWord' noch nicht gefunden wurde
        // ntig, damit spter der Wert hinter dem 'KeyWord=' eingelesen werden kann
        if (!KeyWordFound)
            ReadOk = ReadLine();
    } // continue while there is more to read, we haven't found the keyword and the section didn't end
    while ((ReadOk) && (!KeyWordFound) && (!SectionEnd));

    return KeyWordFound;
}

/*
 * Reads the string after the 'KeyWord=', stops at the end of the line or
 * at the beginning of a comment
 * !! the string is stored in 'ValueString'
 */
void
CIniFile::ReadValueString( void )
{
    char *ValueStringPtr = ValueString;
    const char *LinePtr  = CurrentLine;

    // clear valuestring
    memset( ValueString, ' ', MAXLINELENGTH );
	ValueString[MAXLINELENGTH - 1] = '\0';

    // advance through the strings until we find the end of the source string
    // or the equals character
    while ((*LinePtr != '=') && (*LinePtr != '\0'))
    {
        LinePtr++;
        ValueStringPtr++;
    }

    bool CopyNextChar = true;

    while ( CopyNextChar )
    {
        LinePtr++;
        ValueStringPtr++;

        // we encountered the beginning of a comment, don't continue to copy characters
        if (( *LinePtr == '/') && ((*(LinePtr+1) == '/')) || (*LinePtr == '\0'))
            CopyNextChar = false;
        else
            *ValueStringPtr = *LinePtr;

        // the source line has ended, stop copying characters
        if ( *LinePtr == '\0' )
            CopyNextChar = false;
    }
}

/*
 * Reads an integer value from the currently opened *.ini file
 * returns true if the read was successfull, else false
 */
bool
CIniFile::ReadNumber( const char *SectionName, const char *KeyWord, tSINT32 *ReturnValue )
{
    ASSERT( SectionName != NULL );
    ASSERT( KeyWord != NULL );

    // start at the beginning of the file
    Rewind( );

    if ( FindSection( SectionName ) )
        if ( FindKeyWord( KeyWord ) )
            {

            ReadValueString();
            sscanf( ValueString, "%i", ReturnValue ); // !! don't like 'sscanf()'

            return true;
            }

	*LOG << "Trying to read string '" << KeyWord << "' from section '" <<
			 SectionName << "' in file '" << InputFileName << "' failed!" << std::endl;

    return false;
}

/*
 * Reads an unsigned integer value from the currently opened *.ini file
 * returns true if the read was successfull, else false
 */
bool
CIniFile::ReadNumber( const char *SectionName, const char *KeyWord, tUINT32 *ReturnValue )
{
    ASSERT( SectionName != NULL );
    ASSERT( KeyWord != NULL );

    // start at the beginning of the file
    Rewind( );

    if ( FindSection( SectionName ) )
        if ( FindKeyWord( KeyWord ) )
            {

            ReadValueString();
            sscanf( ValueString, "%u", ReturnValue ); // !! don't like 'sscanf()'

            return true;
            }

	*LOG << "Trying to read string '" << KeyWord << "' from section '" <<
			 SectionName << "' in file '" << InputFileName << "' failed!" << std::endl;

    return false;
}

/*
 * !!! Warning: Extremely dangerous function! Assumes that RetunString already points to
 *              an allocated chunk of memory large enough to hold the string - very bad
 *              Find out a better solution to pass around strings!
 *
 * !!! This function is UGLY! replace asap!
 */
bool
CIniFile::ReadString( const char *SectionName, const char *KeyWord, char *ReturnString )
{
    ASSERT( SectionName != NULL );
    ASSERT( KeyWord != NULL );
    ASSERT( ReturnString != NULL );

    // start at the beginning of the file
    Rewind( );

    if ( FindSection( SectionName ) )
        if ( FindKeyWord( KeyWord ) )
            {

            ReadValueString();

            char *ValueStringPtr = ValueString;
            char *LinePtr  = ReturnString;

            while ( *ValueStringPtr != '"' )
            {
                if ( *ValueStringPtr == '\0' )
                    return false;
                ValueStringPtr++;
            }

            ValueStringPtr++;

            while (( *ValueStringPtr != '"') && ( *ValueStringPtr != '\0' ))
            {
                *LinePtr = *ValueStringPtr;
                ValueStringPtr++;
                LinePtr++;
            }

            *LinePtr = '\0';

            return true;
            }

	*LOG << "Trying to read string '" << KeyWord << "' from section '" <<
			 SectionName << "' in file '" << InputFileName << "' failed!" << std::endl;

    return false;

}

