#ifndef __INIFILE_HPP__
#define __INIFILE_HPP__
#include "General/Configuration.hpp"

/******************************************************************************
 * BlackFish
 * Copyright (C) 1999-2001 Sren Meyer-Eppler, Jutta Jordans
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ******************************************************************************/

/******************************************************************************
    Header file for the IniFile Parser module. The purpose of this module is to
    provide an easy way to read values from a standard DOS-ASCII text file. The
    functions scan the file for certain keywords and return the numbers/text
    belonging to that keyword. The file structure is similar to a windows *.ini
    file: it is divided into paragraphs which makes for easy reading because
    groups of related values can be put together under one topic.
    example:

    Notes:
    [TestWindow]
    XPosition=12    // horizontal window position in pixels (top-left corner)
    YPosition=11
    Width=33
    Height=50
    Name="Test Window"
    TestValue    =    17.765765

    - Keywords must be unique in one paragraph, but it is allowed to have the
     same keyword in different paragraphs, i.e.: Xpos in [TestWindow] and Xpos in
     [HelloWindow]
    - no spaces at the beginning of a line are allowed
    - comments begin with // and cover the rest of the line
    - whitespaces between the keyword the equals sign and the value are allowed
    - keywords must consist of normal characters and are case-sensitive
    - a maximum of MAXLINELENGTH=256 characters is allowed per file line

    (Note: Zur Zeit knnte man genausogut die standard Windows Prozeduren GetPrivate-
    ProfileXXX verwenden, aber erstens ist dies hier ne gute bung, und zweitens
    knnten wir ja spter mal ein komplizierteres Dateiformat verwenden (*.zip z.B.))

    (Note: Ziemlich ineffieziente Implementierung, man sollte eigentlich die relevanten
    Informationen aus der Datei in eine Struktur im Speicher lesen und dann mit
    'ReadInteger()' etc auf den Speicher zugreifen anstatt die Datei jedesmal komplett
    einzulesen...)

    What happens if line length > 256?

    History of Changes:
    - 08.11.1998: file created by BuschnicK
    - 06.02.1999: modified by BuschnicK
    - 21.02.1999: polished by BuschnicK, notes:
                  - nur sehr unvollstndig getestet
                  - es fehlt noch die Mglichkeit strings,tFLOATs etc zu lesen
    - 25.02.1999: modified by BuschnicK according to design rules in
                  'Large Scale C++ Software Design' (Lakos)
    - 27.02.1999: updated variable declarations according to the new names in
                  type.hpp
    - 21.03.1999: !! external include guards can be replaced by '#pragma once'
******************************************************************************/

#pragma warning ( push, 1 )
#include <fstream>
#pragma warning ( pop )

#ifndef __TYPE_HPP__ // redundant (external) include guard (Lakos)
#include "general\type.hpp"
#endif

class
CIniFile
{
public:

    CIniFile( void );
    CIniFile( const char *FileName );

    ~CIniFile( void );

    void Open( const char *FileName );
    void Close( void );

    bool ReadNumber( const char *SectionName, const char *KeyWord, tSINT32 *ReturnValue );
    bool ReadNumber( const char *SectionName, const char *KeyWord, tUINT32 *ReturnValue );

    bool ReadString( const char *SectionName, const char *KeyWord, char *ReturnString );  // !!! beware this function!

private:

	std::ifstream *InputFile;
	char *InputFileName;

    enum { MAXLINELENGTH = 256 };   // !! right way to do this? a constant with object scope is what I want
    char CurrentLine[MAXLINELENGTH];
    char ValueString[MAXLINELENGTH]; // !! besser nur als tempvariable in den funktionen?

    void Initialize( void );

    void Rewind( void );
    bool ReadLine( void );
    bool FindSection( const char *SectionName );
    bool FindKeyWord( const char *KeyWord );
    bool FindString( const char *StringName );
    void ReadValueString( void );

	// copy and assignment is not allowed
    CIniFile( const CIniFile &IniFile );
    CIniFile & operator = ( const CIniFile &IniFile );
};

#endif
