#include "Map.hpp"

/******************************************************************************
 * BlackFish
 * Copyright (C) 1999-2001 Sren Meyer-Eppler, Jutta Jordans
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ******************************************************************************/


#pragma warning ( push, 1 )
#include <fstream>
#include <algorithm>
#pragma warning ( pop )

#include "general\vector.hpp"
#include "BlackFish\Sprite.hpp"

using std::list;
using std::vector;
using std::make_heap;
using std::sort_heap;
using std::pop_heap;
using std::find;


/**
 * default constructor, make empty map
 */
CMap::CMap( void )
    : Cell( NULL ),
	  FileName( )
{
	Header.Version	= DATAVERSION;
	Header.Width	= 32;
	Header.Height	= 32;
	memset( Header.Name, ' ', sizeof( Header.Name ));
	memcpy( Header.Name, "default map", 12 );
	memset( Header.Author, ' ', sizeof ( Header.Author ));
	memcpy( Header.Author, "system generated", 17 );
	memset( Header.Padding, 0, sizeof ( Header.Padding ));

    Cell = new CBaseCell[ static_cast<tUINT32>( Header.Width * Header.Height )];
    ASSERT( Cell != NULL );
}

/*
 * constructor to make empty map of certain width and height
 */
CMap::CMap ( const tMAPHEADER & MapHeader )
	: Cell( NULL ),
	  FileName( )
{
	memcpy( &Header, &MapHeader, sizeof( MapHeader ));

	// Padding is not allowed to hold any information, so overwrite anything that
	// might be there - saves the caller from having to this himself
	memset( Header.Padding, 0, sizeof ( Header.Padding ));

    Cell = new CBaseCell[static_cast<tUINT32>( Header.Width * Header.Height )];
    ASSERT( Cell != NULL );
}

/*
 * constructor to make Map from File
 */
CMap::CMap( const char * NewFileName )
	: Cell( NULL ),
	  FileName( NewFileName )
{
    Load( NewFileName );
}

/*
 * destructor, free array of cells
 */
CMap::~CMap( )
{
	DeleteCells();
}

/*
 *	Delete the entire array of cells, to be called by destructor and
 *  before loading new map
 */
void
CMap::DeleteCells( void )
{
    SAFEDELETEARRAY( Cell ); // delete topmost layer (the groundtiles) calls CBaseCell destructor which
							 // deletes all layers
}

/*
 */
void
CMap::InsertSprite( CVector *Position, CSprite *Sprite )
{
	ASSERT( Position != NULL );
	ASSERT( Sprite != NULL );

    std::list<CSprite*> & List = *GetSpriteList( 
		static_cast<tSINT32>( Position->x / 32.0 ),
		static_cast<tSINT32>( Position->y / 32.0 ));

	List.push_back( Sprite );
	List.sort( SpriteComparer );
}

/*
 *
 */
void
CMap::RemoveSprite( CVector *Position, CSprite *Sprite )
{
	ASSERT( Position != NULL );
	ASSERT( Sprite != NULL );

	GetSpriteList( static_cast<tSINT32>( Position->x / 32.0 ),
				   static_cast<tSINT32>( Position->y / 32.0 ))->remove( Sprite );
}

/*
 * scans the map for unused layers and removes these (unused = tileset 0 tile 0)
 */
void
CMap::Optimize( void )
{
    CCellinfo *OldCheckTile;
    CCellinfo *CheckTile;

    for ( tSINT32 j = 0; j < Header.Height; j++ )
        for ( tSINT32 i = 0; i < Header.Width; i++ )
        {
            CheckTile = GetFirstLayer( i, j );
			ASSERT( CheckTile != NULL );

            while ( CheckTile->NextLayer != NULL )
            {
                OldCheckTile = CheckTile;
                CheckTile = CheckTile->NextLayer;
                if (( CheckTile->TileNr == 2 ) && ( CheckTile->TileSet == 0 ))
                {
                    OldCheckTile->NextLayer = CheckTile->NextLayer;
                    SAFEDELETE ( CheckTile );
                    CheckTile = OldCheckTile;
				}
			}
		}
}

/*
 * save the map
 * !! currently doesn't include sprites
 */
void
CMap::Save( const char *FileName )
{
	ASSERT( FileName != NULL );

	// make output file and open as binary
	std::ofstream MapFile( FileName, std::ios::out|std::ios::binary );
	ASSERT( MapFile.good( ));

	// header is supposed to be 256 bytes
	MapFile.write( (char *)&Header, sizeof( Header ));

	// temporary variable used to write to file
	tSINT32 Data = 0;
	tUINT32 uData = 0;

	for ( tSINT32 j = 0; j < Header.Height; j++ )
		for ( tSINT32 i = 0; i < Header.Width; i++ )
		{
			// write the base layer's sprite info
			Data = GetFirstLayer( i, j )->TileSet;
			MapFile.write( (char *)&Data, sizeof( Data ));
			Data = GetFirstLayer( i, j )->TileNr;
			MapFile.write( (char *)&Data, sizeof( Data ));
 			Data = GetBaseCell( i, j )->GetFlags( );
			MapFile.write( (char *)&Data, sizeof( Data ));
            uData = GetBaseCell( i, j )->GetFlags( );
			MapFile.write( (char *)&uData, sizeof( uData ));
            Data = GetBaseCell( i, j )->GetWall( ) != NULL; // used as a bool here
			MapFile.write( (char *)&Data, sizeof( Data ));

			// get number of top layers
			tSINT32 NrOfLayers = 0;
            CCellinfo *Iterator = GetFirstLayer( i, j );
			while ( Iterator->NextLayer != NULL )
			{
				NrOfLayers++;
				Iterator = Iterator->NextLayer;
			}

			// write number of top layers
			Data = NrOfLayers;
			MapFile.write( (char *)&Data, sizeof( Data ));

			Iterator = GetFirstLayer( i, j );
			for ( tSINT32 k = 0; k < NrOfLayers; k++ )
			{
				Iterator = Iterator->NextLayer;
				Data = Iterator->TileSet;
				MapFile.write( (char *)&Data, sizeof( Data ));
				Data = Iterator->TileNr;
				MapFile.write( (char *)&Data, sizeof( Data ));
			}

			if ( GetBaseCell( i, j )->GetWall( ) != NULL )
			{
				Data = GetBaseCell( i, j )->GetWall( )->TileSet;
				MapFile.write( (char *)&Data, sizeof( Data ));
				Data = GetBaseCell( i, j )->GetWall( )->TileNr;
				MapFile.write( (char *)&Data, sizeof( Data ));
			}
		}
	MapFile.close( );
}

/*
 * Load a new map from file (without making new map object)
 * !! currently doesn't support sprites
 */
void
CMap::Load( const char* NewFileName )
{
	ASSERT( NewFileName != NULL );

	// delete old array of cells:
	DeleteCells( );

	// make output file and open as binary
	std::ifstream MapFile( NewFileName, std::ios::in|std::ios::binary );
	ASSERT( MapFile.good( ));

	// header is supposed to be 256 bytes
	MapFile.read( (char *)&Header, sizeof( Header ));
	if ( Header.Version != DATAVERSION )
	{
		*LOG << "Loading of '" << NewFileName << "' failed, expecting version "
			 << DATAVERSION << ", found " << Header.Version << "." << std::endl;
		QUIT
	}

	ASSERT( Cell == NULL );
    Cell = new CBaseCell[ static_cast<tUINT32>( Header.Width * Header.Height )];
    ASSERT( Cell != NULL );

	tSINT32 Data1 = 0, Data2 = 0;
	tUINT32 uData = 0;

	for ( tSINT32 j = 0; j < Header.Height; j++ )
		for ( tSINT32 i = 0; i < Header.Width; i++ )
		{
			// read the base layer's sprite info
            bool WallFlag = false;

			MapFile.read( (char *)&Data1, sizeof( Data1 ));
			MapFile.read( (char *)&Data2, sizeof( Data2 ));
			SetFirstLayer( i, j, Data1, Data2, NULL );
			MapFile.read( (char *)&uData, sizeof( uData ));
			GetBaseCell( i, j )->SetFlags( uData );

			MapFile.read( (char *)&Data1, sizeof( Data1 ));
			WallFlag = Data1 != 0;

			// get number of top layers
			tSINT32 NrOfLayers = 0;
			MapFile.read( (char *)&Data1, sizeof( Data1 ));
			NrOfLayers = Data1;

            CCellinfo *Iterator = GetFirstLayer( i, j );
			for ( tSINT32 k = 0; k < NrOfLayers; k++ )
			{
                Iterator->NextLayer = new CCellinfo( );
				ASSERT( Iterator->NextLayer != NULL );
				Iterator = Iterator->NextLayer;

				MapFile.read( (char *)&Data1, sizeof( Data1 ));
				Iterator->TileSet = Data1;
				MapFile.read( (char *)&Data1, sizeof( Data1 ));
				Iterator->TileNr = Data1;
			}

			if ( WallFlag )
			{
				MapFile.read( (char *)&Data1, sizeof( Data1 ));
				MapFile.read( (char *)&Data2, sizeof( Data2 ));
				SetWall( i, j, Data1, Data2 );
				ASSERT( GetBaseCell( i, j )->GetWall( ) != NULL );
			}
		}
	MapFile.close( );

	FileName = NewFileName;

*LOG << std::endl << NewFileName << " read successfully:" << std::endl
	 << "    Name:      " << Header.Name << std::endl
	 << "    Author:    " << Header.Author << std::endl
	 << "    Dimension: " << Header.Width << "x" << Header.Height << std::endl;
}

/*
 *
 */
tSINT32
CMap::GetHeuristic( tSINT32 x1, tSINT32 y1, tSINT32 x2, tSINT32 y2 )
{
	tSINT32 abs1 = x2 - x1 > 0 ? x2 - x1 : x1 - x2;
	tSINT32 abs2 = y2 - y1 > 0 ? y2 - y1 : y1 - y2;
	tSINT32 Result = abs1 < abs2 ? abs1 * COST_DIAGONAL : abs2 * COST_DIAGONAL; // min(|dx|,|dy|)
	Result += abs1 > abs2 ? (abs1 - abs2) * COST_ALIGNED : (abs2 - abs1) * COST_ALIGNED; // ||dx|-|dy||
	return Result;
}

/*
 *
 */
tSINT32
CMap::GetCost( tSINT32 x1, tSINT32 y1, tSINT32 x2, tSINT32 y2 )
{
	if (( x1 - x2 ) && ( y1 - y2 )) // diagonal movement
		return COST_DIAGONAL;

	return COST_ALIGNED;
}

/*
 *
 */
void
CMap::DeleteNodes( vector< CNode * > &Open, vector< CNode * > &Closed )
{
    std::vector< CNode * >::size_type i;
	for ( i = 0; i < Open.size( ); i++ )
		SAFEDELETE( Open[i] );
	Open.clear( );

	for ( i = 0; i < Closed.size( ); i++ )
		SAFEDELETE( Closed[i] );
	Closed.clear( );
}

/*
 *
 */
void
CMap::AddNode( vector< CNode * > &Open, vector< CNode * > &Closed, CNode *Node )
{
    std::vector< CNode * >::iterator i;
	for ( i=Closed.begin( ); i != Closed.end( ); i++ )
	{
		if (( Node->x == (*i)->x ) && ( Node->y == (*i)->y ))
			if ( (*i)->Cost < Node->Cost )
			{
				SAFEDELETE( Node );
				return;
			}
			else
			{
				SAFEDELETE( *i );
				Closed.erase( i );
				break;
			}
	}

	for ( i=Open.begin( ); i != Open.end( ); i++ )
	{
		if (( Node->x == (*i)->x ) && ( Node->y == (*i)->y ))
			if ( (*i)->Cost < Node->Cost )
			{
				SAFEDELETE( Node );
				return;
			}
			else
			{
				SAFEDELETE( *i );
				Open.erase( i );
				break;
			}
	}

	Open.push_back( Node );
}

bool greater( const CNode *a, const CNode *b ); // forward declaration, used in FindPath to sort CNodeS *

/*
 * !! not at all optimized:
 * - find a way to reduce allocations/deallocations
 * - find a structure that gives efficient results for searches with two different search keys,
 *   i.e. the open list must be sorted by coordinate (in addnode) and by heuristic (findpath)
 * @return true, if a path could be found, false otherwise
 */
bool
CMap::FindPath( const CVector & PointA, const CVector & PointB, list<CVector> *Path )
{
    vector< CNode * > Open;
    vector< CNode * > Closed;

	tSINT32 Iteration = 0;

	tSINT32 X	  = static_cast<tSINT32>(PointA.x / 32.0f);
	tSINT32 Y	  = static_cast<tSINT32>(PointA.y / 32.0f);
	tSINT32 GoalX = static_cast<tSINT32>(PointB.x / 32.0f);
	tSINT32 GoalY = static_cast<tSINT32>(PointB.y / 32.0f);

	// early exit if goal is impassible. Maybe calculating a route to a spot
	// near the goal might be more appropriate - on the other hand it would be more complicated
	// and I can foresee some situations in which the results would be unexpected/undesired
	if ( IsBlocking( GoalX, GoalY ))
        return false;

	// early exit if start and destination are the same
	if ( X == GoalX && Y == GoalY )
        return true;

    Open.push_back( new CNode( X, Y, 0, GetHeuristic( X, Y, GoalX, GoalY )));
	make_heap( Open.begin( ), Open.end( ), greater );

	while (( !Open.empty( )) && (++Iteration <= MAXITERATIONS))
	{
		pop_heap( Open.begin( ), Open.end( ), greater ); // restructure heap, so smallest element is last in the vector

        CNode *Current = Open[Open.size() - 1];

		Open.pop_back( );	// remove the last element (the smallest) from the vector

		X = Current->x;
		Y = Current->y;

		if (( X == GoalX ) && ( Y == GoalY ))	// we arrived at the goal node
		{
            CNode *Iterator = Current;
			do	// construct path
			{
                Path->push_front( CVector( static_cast<tFLOAT>( Iterator->x * 32 ), static_cast<tFLOAT>( Iterator->y * 32 ), 0.0f ));
				Iterator = Iterator->Parent;
			}
			while ( Iterator != NULL );

			SAFEDELETE( Current );
			DeleteNodes( Open, Closed );
            return true;
		}

		tSINT32 NewX, NewY;
		NewX = X-1; NewY = Y-1;
		if ( !IsBlocking( X, Y, NewX, NewY ))
            AddNode( Open, Closed, new CNode( NewX, NewY,
				Current->Cost + GetCost( X, Y, NewX, NewY ),
				GetHeuristic( NewX, NewY, GoalX, GoalY ), Current ));

		NewX = X; NewY = Y-1;
		if ( !IsBlocking( X, Y, NewX, NewY ))
            AddNode( Open, Closed, new CNode( NewX, NewY,
				Current->Cost + GetCost( X, Y, NewX, NewY ),
				GetHeuristic( NewX, NewY, GoalX, GoalY ), Current ));

		NewX = X+1; NewY = Y-1;
		if ( !IsBlocking( X, Y, NewX, NewY ))
            AddNode( Open, Closed, new CNode( NewX, NewY,
				Current->Cost + GetCost( X, Y, NewX, NewY ),
				GetHeuristic( NewX, NewY, GoalX, GoalY ), Current ));

		NewX = X+1; NewY = Y;
		if ( !IsBlocking( X, Y, NewX, NewY ))
            AddNode( Open, Closed, new CNode( NewX, NewY,
				Current->Cost + GetCost( X, Y, NewX, NewY ),
				GetHeuristic( NewX, NewY, GoalX, GoalY ), Current ));

		NewX = X+1; NewY = Y+1;
		if ( !IsBlocking( X, Y, NewX, NewY ))
            AddNode( Open, Closed, new CNode( NewX, NewY,
				Current->Cost + GetCost( X, Y, NewX, NewY ),
				GetHeuristic( NewX, NewY, GoalX, GoalY ), Current ));

		NewX = X; NewY = Y+1;
		if ( !IsBlocking( X, Y, NewX, NewY ))
            AddNode( Open, Closed, new CNode( NewX, NewY,
				Current->Cost + GetCost( X, Y, NewX, NewY ),
				GetHeuristic( NewX, NewY, GoalX, GoalY ), Current ));

		NewX = X-1; NewY = Y+1;
		if ( !IsBlocking( X, Y, NewX, NewY ))
            AddNode( Open, Closed, new CNode( NewX, NewY,
				Current->Cost + GetCost( X, Y, NewX, NewY ),
				GetHeuristic( NewX, NewY, GoalX, GoalY ), Current ));

		NewX = X-1; NewY = Y;
		if ( !IsBlocking( X, Y, NewX, NewY ))
            AddNode( Open, Closed, new CNode( NewX, NewY,
				Current->Cost + GetCost( X, Y, NewX, NewY ),
				GetHeuristic( NewX, NewY, GoalX, GoalY ), Current ));

		make_heap( Open.begin( ), Open.end( ), greater );
		Closed.push_back( Current );
	}

	DeleteNodes( Open, Closed );
    return false;
}

void
CMap::SetName(const char* MapName)
{
	memcpy(Header.Name, MapName, 64);
}

void
CMap::SetAuthor(const char* Author)
{
	memcpy(Header.Author, Author, 64);
}

const char*
CMap::GetName()
{
	return Header.Name;
}

const char*
CMap::GetAuthor()
{
	return Header.Author;
}
//////////////////////////////////////////////////////////////////////
bool operator > ( const CNode &a, const CNode &b )   // object comparison
{
    // returns true if 'a' has a better 'f' (=cost to reach the node + heuristic) than 'b'
    // if nodes have equal 'f' returns true if 'a' has a better heuristic than 'b'
	// this makes the A* algorithm always chase down the nodes, which are already closest
	// to the goal first
    return a.Total > b.Total ? true : ( a.Total == b.Total ? a.Heuristic > b.Heuristic : false );
}

bool greater( const CNode *a, const CNode *b )       // pointer comparison
{
	return *a > *b;
}
//////////////////////////////////////////////////////////////////////
/*
 *
 */
CBaseCell::CBaseCell( )
	: FirstLayer( ),
	  Wall( NULL ),
	  Flags( 0 ),
	  SpriteList( )
{
}

/*
 *
 */
CBaseCell::~CBaseCell( )
{
	DeleteAllLayers( );
	SAFEDELETE( Wall );
}

/*
 *
 */
bool
CBaseCell::IsBlocking( void )
{
	// check whether the cell itself blocks
    if (( Flags & CMap::FLAGBLOCKING ) != 0 )
        return true;

	// check for blocking sprites occupying this cell
    std::list< CSprite * >::iterator i = SpriteList.begin( );
	while ( i != SpriteList.end( ))
	{
        if (((*i)->GetFlags( ) & CSprite::FLAGBLOCKING ) != 0 )
            return true;
		++i;
	}

    return false;
}
